import re
import os
import fnmatch
from pyaerocom import const, logger
from pyaerocom._lowlevel_helpers import BrowseDict
from pyaerocom.exceptions import DataSearchError

class AerocomBrowser(BrowseDict):
    """Interface for browsing all Aerocom data direcories

    Note
    ----
    Use :func:`browse` to find directories matching a
    certain search pattern.
    The class methods :func:`find_matches` and :func:`find_dir` both use
    :func:`browse`, the only difference is, that the :func:`find_matches` adds
    the search result (a list with strings) to

    """
    def _browse(self, name_or_pattern, ignorecase=True, return_if_match=True):
        """Search all Aerocom data directories that match input name or pattern

        Note
        ----
        Please do not use this function but either
        Parameters
        ----------
        name_or_pattern : str
            name or pattern of data (can be model or obs data)
        ignorecase : bool
            if True, upper / lower case is ignored
        return_if_match : bool
            if True, then the data directory is returned as string, if it can
            be found, else, only a list is returned that contains all
            matches. The latter takes longer since the whole database is
            searched.

        Returns
        -------
        :obj:`str` or :obj:`list`
            Data directory (str, if ``return_if_match`` is True) or list
            containing valid Aerocom names (which can then be used to
            retrieve the paths)

        Raises
        ------
        DataSearchError
            if no match or no unique match can be found
        """
        pattern = fnmatch.translate(name_or_pattern)
        _candidates = []
        _msgs = []
        _warnings = []

        for obs_id, obs_path in const.OBSLOCS_UNGRIDDED.items():
            if ignorecase:
                match = name_or_pattern.lower() == obs_id.lower()
            else:
                match = name_or_pattern == obs_id
            if match:
                logger.info("Found match for search pattern in obs network "
                            "directories {}".format(obs_id))
                path = os.path.normpath(obs_path)
                if os.path.exists(path):
                    self[obs_id] = path
                    if return_if_match:
                        return path
            else:
                if ignorecase:
                    match = bool(re.search(pattern, obs_id, re.IGNORECASE))
                else:
                    match = bool(re.search(pattern, obs_id))
                if match:
                    path = os.path.normpath(obs_path)
                    if os.path.exists(path):
                        self[obs_id] = path
                        _candidates.append(obs_id)
                        if return_if_match:
                            return path

        for search_dir in const.DATA_SEARCH_DIRS:
            # get the directories
            if os.path.isdir(search_dir):
                #subdirs = listdir(search_dir)
                subdirs = [x for x in os.listdir(search_dir) if
                           os.path.isdir(os.path.join(search_dir, x))]
                for subdir in subdirs:
                    if ignorecase:
                        match = bool(re.search(pattern, subdir,re.IGNORECASE))
                    else:
                        match = bool(re.search(pattern, subdir))
                    if match:
                        _dir = os.path.normpath(os.path.join(search_dir, subdir))
                        _rnsubdir = os.path.join(_dir, "renamed")
                        if os.path.isdir(_rnsubdir):
                            logger.info("{} has subdir renamed. Using that one"
                                        .format(_dir))
                            _dir = _rnsubdir
                        if any([_dir in x for x in self.values()]):
                            # directory was already found before
                            continue
                        # append name of candidate ...
                        _candidates.append(subdir)
                        # ... and the corresponding data directory
                        self[subdir] = _dir

                        # now check if it is actually an exact match, if
                        # applicable
                        if return_if_match:

                            if ignorecase:
                                match = name_or_pattern.lower() == subdir.lower()
                            else:
                                match = name_or_pattern == subdir
                            if match:
                                logger.info("Found match for ID {}"
                                            .format(name_or_pattern))
                                if return_if_match:
                                    return _dir

            else:
                _msgs.append('directory %s does not exist\n'
                                 %search_dir)
        for msg in _msgs:
            logger.info(msg)

        for warning in _warnings:
            logger.warning(warning)

        if len(_candidates) == 0:
            raise DataSearchError('No matches could be found for search pattern '
                          '{}'.format(name_or_pattern))
        if return_if_match:
            if len(_candidates) == 1:
                logger.info("Found exactly one match for search pattern "
                            "{}: {}".format(name_or_pattern, _candidates[0]))
                return self[_candidates[0]]
            raise DataSearchError('Found multiple matches for search pattern {}. '
                          'Please choose from {}'.format(name_or_pattern,
                                              _candidates))
        return _candidates

    @property
    def dirs_found(self):
        """All directories that were found"""
        return list(self.values())

    @property
    def ids_found(self):
        """All data IDs that were found"""
        return list(self.keys())

    def find_data_dir(self, name_or_pattern, ignorecase=True):
        """Find match of input name or pattern in Aerocom database

        Parameters
        ----------
        name_or_pattern : str
            name or pattern of data (can be model or obs data)
        ignorecase : bool
            if True, upper / lower case is ignored

        Returns
        -------
        str
            data directory of match

        Raises
        ------
        DataSearchError
            if no matches or no unique match can be found
        """
        if name_or_pattern in self:
            logger.info('{} found in instance of AerocomBrowser'.format(name_or_pattern))
            return self[name_or_pattern]
        logger.info('Searching database for {}'.format(name_or_pattern))
        return self._browse(name_or_pattern, ignorecase=ignorecase,
                            return_if_match=True) #returns list

    def find_matches(self, name_or_pattern, ignorecase=True):
        """Search all Aerocom data directories that match input name or pattern

        Parameters
        ----------
        name_or_pattern : str
            name or pattern of data (can be model or obs data)
        ignorecase : bool
            if True, upper / lower case is ignored

        Returns
        -------
        list
            list of names that match the pattern (corresponding paths can be
            accessed from this class instance)

        Raises
        ------
        DataSearchError
            if no matches can be found
        """
        return self._browse(name_or_pattern, ignorecase=ignorecase,
                            return_if_match=False) #returns list

if __name__ == "__main__":
    browser = AerocomBrowser()
    dd = browser.find_data_dir('TM5_AP3-CTRL2016*')
    ea = browser.find_data_dir('Earli*')
    ea1 = browser.find_matches('Earlinet')

    print(ea)
    print(ea1)
# =============================================================================
#     try:
#         data_dir = browser.find_data_dir('*Cam5.3-Oslo*')
#     except DataSearchError as e:
#         print(repr(e))
#
#
#
#     matches = browser.find_matches('*Cam5.3-Oslo*')
#
#     for match in matches:
#         print('{}: {}'.format(match, browser[match]))
#
#     data_dir_earlinet = browser.find_data_dir('EARLIN*')
#
#     data_dir_earlinet = browser.find_data_dir('EARLIN*')
#
#     browser.find_data_dir('Earlinet')
#
#
#
#
# =============================================================================



from pyaerocom.mathutils import is_strictly_monotonic, make_binlist
from pyaerocom.variable_helpers import get_variable



class VarinfoWeb:
    """
    Additional variable information relevant for AeroVal web output

    Attributes
    ----------
    var_name : str
        Name of variable (AeroCom name, not web display name)
    cmap_bins : list
        Value bins for web display
    cmap : str
        name of colormap for web display

    Parameters
    ----------
    var_name : str
        Name of variable (AeroCom name, not web display name)
    cmap : str, optional
        name of colormap for web display. If None, the
        colormap associated with the input variable is used (via
        :func:`pyaerocom.variable.Variable.get_cmap`). Defaults to None.
    cmap_bins : list, optional
        Value bins for web display. If None, then they are inferred from input
        vmin and vmax, or, if the latter are also None,
        from attrs :attr:`pyaerocom.variable.Variable.minimum` and
        :attr:`pyaerocom.variable.Variable.maximum`. If the latter are not
        defined an AttributeError will be thrown on initialisation.
    vmin : float, optional
        lower end of range
    vmax : float, optional
        upper end of range
    """
    _num_bins = 8

    def __init__(self, var_name: str, cmap: str = None,
                 cmap_bins: list = None, vmin: float = None,
                 vmax: float = None):
        if cmap_bins is not None:
            if vmin is not None or vmax is not None:
                raise ValueError('please provide either vmin and vmax OR '
                                 'cmap_bins, not both...')
            if not is_strictly_monotonic(cmap_bins):
                raise ValueError('cmap_bins need to be strictly monotonic')

        self.var_name = var_name
        self.cmap_bins = cmap_bins
        self.cmap = cmap
        self.autofill_missing(vmin, vmax)

    @property
    def vmin(self) -> float:
        """
        Lower end of range
        """
        return self.cmap_bins[0]

    @property
    def vmax(self) -> float:
        """
        Upper end of range
        """
        return self.cmap_bins[-1]

    def autofill_missing(self, vmin:float=None, vmax:float=None) -> None:
        """
        Autofill missing attributes related to cmap bins and cmap

        Parameters
        ----------
        vmin : float, optional
            lower end of range
        vmax : float, optional
            upper end of range

        Returns
        -------
        None

        """
        var = None
        if vmin is not None and vmax is not None:
            self.cmap_bins = make_binlist(vmin, vmax, self._num_bins)
        elif self.cmap_bins is None:
            var = get_variable(self.var_name)
            self.cmap_bins = var.get_cmap_bins(infer_if_missing=True)

        if self.cmap is None:
            if var is None:
                var = get_variable(self.var_name)
            self.cmap = var.get_cmap()

    @staticmethod
    def from_dict(dict):
        """
        Instantiate from dictionary

        Parameters
        ----------
        dict : dict
            settings

        Returns
        -------
        VarinfoWeb
            instantiated instance
        """
        return VarinfoWeb(**dict)

    def to_dict(self):
        """
        Convert to dictionary

        Returns
        -------
        dict
        """
        dd = {**self.__dict__}
        dd['vmin'] = self.vmin
        dd['vmax'] = self.vmax
        return dd



from ._init_helpers import (_init_supplemental, _init_logger,
                            LOGLEVELS, change_verbosity)

__version__, __dir__ = _init_supplemental()

logger, print_log = _init_logger()

from .config import Config
# Instantiate default configuration
const = Config()

# Sub-packages
from . import io
from . import plot
from . import tools
from . import scripts


# Imports
from . import obs_io
from . import metastandards
from . import vertical_profile
from . import mathutils
from . import geodesy
from . import region_defs
from . import region
#from . import vert_coords
from . import stationdata
from . import griddeddata
from . import ungriddeddata
from . import colocation
from . import var_groups
from . import combine_vardata_ungridded
from . import helpers_landsea_masks
from . import helpers
from . import trends_helpers

from . import trends_engine


# custom toplevel classes
from .variable import Variable
from .region import Region
from .vertical_profile import VerticalProfile
from .stationdata import StationData
from .griddeddata import GriddedData
from .ungriddeddata import UngriddedData
from .filter import Filter
from .colocateddata import ColocatedData
from .colocation_auto import ColocationSetup, Colocator
from .tstype import TsType
from .time_resampler import TimeResampler
from .io.helpers import search_data_dir_aerocom
from .variable_helpers import get_variable
from .utils import create_varinfo_table
from .access_testdata import initialise as initialise_testdata

from . import aeroval

# toplevel functions
from pyaerocom.tools import browse_database



################################################################
# read/__init__.py
#
# init for data reading
#
# this file is part of the aerocom_pt package
#
#################################################################
# Created 20171030 by Jan Griesfeller for Met Norway
#
# Last changed: See git log
#################################################################

#Copyright (C) 2017 met.no
#Contact information:
#Norwegian Meteorological Institute
#Box 43 Blindern
#0313 OSLO
#NORWAY
#E-mail: jan.griesfeller@met.no
#This program is free software; you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation; either version 3 of the License, or
#(at your option) any later version.
#This program is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#GNU General Public License for more details.
#You should have received a copy of the GNU General Public License
#along with this program; if not, write to the Free Software
#Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#MA 02110-1301, USA

# =============================================================================
# from .read_aeronet_sdav2 import ReadAeronetSDAV2
def geopy_available():
    """Helper method that checks if geopy library is available

    Required for import of ReadAeolusL2aData

    Returns
    -------
    bool
        True, if library is available, else False
    """
    try:
        import geopy
        return True
    except ModuleNotFoundError:
        from logging import getLogger
        logger = getLogger('pyaerocom')
        logger.warning('geopy library is not available. Aeolus data read not '
                       'enabled')
    return False

def coda_available():
    """Helper method that checks if coda library is available

    Required for import of ReadAeolusL2aData and ReadSentinel5pL2Data

    Returns
    -------
    bool
        True, if library is available, else False
    """
    try:
        import coda
        return True
    except ModuleNotFoundError:
        from logging import getLogger
        logger = getLogger('pyaerocom')
        logger.warning('coda library is not available. Sentinel5P and Aeolus data read not enabled')
    return False

from .aerocom_browser import AerocomBrowser
from .readungriddedbase import ReadUngriddedBase

# low level EBAS I/O routines
from .ebas_nasa_ames import EbasNasaAmesFile
from .ebas_file_index import EbasSQLRequest, EbasFileIndex

# Pyaerocom reading interface classes
from .read_aeronet_invv2 import ReadAeronetInvV2
from .read_aeronet_invv3 import ReadAeronetInvV3
from .read_aeronet_sdav2 import ReadAeronetSdaV2
from .read_aeronet_sdav3 import ReadAeronetSdaV3
from .read_aeronet_sunv2 import ReadAeronetSunV2
from .read_aeronet_sunv3 import ReadAeronetSunV3
from .read_earlinet import ReadEarlinet
from .read_ebas import ReadEbas
from .read_gaw import ReadGAW
from .read_aasetal import ReadAasEtal
from .read_ghost import ReadGhost
from .read_airnow import ReadAirNow
from .read_marcopolo import ReadMarcoPolo
from .read_eea_aqerep import ReadEEAAQEREP
from .read_eea_aqerep_v2 import ReadEEAAQEREP_V2

from .readgridded import ReadGridded
from .read_mscw_ctm import ReadMscwCtm
from .readungridded import ReadUngridded
from .fileconventions import FileConventionRead

if geopy_available() and coda_available():
    # the coda and geopy libraries are needed to read l2 data of the supported satellites
    # Aeolus and Sentinel5P
    from .read_aeolus_l2a_data import ReadL2Data
    from .read_sentinel5p_data import ReadL2Data

from . import helpers_units



"""Interface for reading EEA AqERep files (formerly known as Airbase data).

This file is part of the pyaerocom package.

#################################################################
# Created 20120128 by Jan Griesfeller for Met Norway
#
# Last changed: See git log
#################################################################

# Copyright (C) 2021 met.no
# Contact information:
# Norwegian Meteorological Institute
# Box 43 Blindern
# 0313 OSLO
# NORWAY
# E-mail: jan.griesfeller@met.no
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA 02110-1301, USA

Example
-------
look at the end of the file
"""

from pyaerocom import const
from pyaerocom.io.read_eea_aqerep_base import ReadEEAAQEREPBase


class ReadEEAAQEREP_V2(ReadEEAAQEREPBase):
    """Class for reading EEA AQErep data

    Extended class derived from  low-level base class :class: ReadUngriddedBase
    that contains the main functionality.
    """
    #: Name of the dataset (OBS_ID)
    DATA_ID = const.EEA_V2_NAME

    #: List of all datasets supported by this interface
    SUPPORTED_DATASETS = [DATA_ID]

    #: Eionet offers 2 data revisions
    #: E2a (near real time) and E1a (quality controlled)
    #: this class reads the E2a data for now.
    # But by changing the base path
    # and this constant, it can also read the E1a data set
    DATA_PRODUCT = 'E1a'

    #: file masks for the data files
...


if __name__ == "__main__":

    # Test that the reading routine works
    import getpass
    username = getpass.getuser()
    import pyaerocom as pya

    if username == 'jang':
        from pyaerocom.io.read_eea_aqerep_v2 import ReadEEAAQEREP_V2
        pya.change_verbosity(new_level='warning', log=pya.logger)

        # filename = '/lustre/storeA/project/aerocom/aerocom1/AEROCOM_OBSDATA/EEA_AQeRep.v2/download/AT/AT_5_48881_2019_timeseries.csv.gz'
        # r = ReadEEAAQEREP_V2()
        # data = r.read_file(filename, 'concpm10')
        #
        # filename = '/lustre/storeA/project/aerocom/aerocom1/AEROCOM_OBSDATA/EEA_AQeRep.v2/download/AT/AT_5_48900_2019_timeseries.csv.gz'
        # r = ReadEEAAQEREP_V2()
        # data2 = r.read_file(filename, 'concpm10')


        import logging
        # limit the data read for testing
        # ReadEEAAQEREP_V2.FILE_MASKS['concso2'] = '**/AT*_1_*_timeseries.csv*'
        # ReadEEAAQEREP_V2.FILE_MASKS['concpm10'] = '**/XK*_5_*_timeseries.csv*'
        # ReadEEAAQEREP_V2.FILE_MASKS['conco3'] = '**/XK*_7_*_timeseries.csv*'
        # ReadEEAAQEREP_V2.FILE_MASKS['vmro3'] = '**/XK*_7_*_timeseries.csv*'
        # ReadEEAAQEREP_V2.FILE_MASKS['concno2'] = '**/XK*_8_*_timeseries.csv*'
        # ReadEEAAQEREP_V2.FILE_MASKS['concno2'] = '**/AT*_8_*_timeseries.csv*'
        # ReadEEAAQEREP_V2.FILE_MASKS['concco'] = '**/AT*_10_*_timeseries.csv*'
        # ReadEEAAQEREP_V2.FILE_MASKS['concno'] = '**/AT*_38_*_timeseries.csv*'
        # ReadEEAAQEREP_V2.FILE_MASKS['concpm25'] = '**/XK*_6001_*_timeseries.csv*'
        station_id = {}
        station_id['concpm25'] = 'XK0002A'
        station_id['vmro3'] = 'XK0002A'
        station_id['concno2'] = 'AT31703'
        station_id['concpm10'] = 'XK0001A'
        # station_id['conco3'] = 'XK0002A'
        # station_id['concno2'] = 'XK0002A'
        # station_id['concco'] = 'XK0002A'
        # station_id['concco'] = 'AT4S416'
        # station_id['concno'] = 'XK0002A'
        # station_id['concno'] = 'AT4S416'
        # station_id['concso2'] = 'AT31703'

        var_names_to_test = station_id.keys()
        # var_names_to_test = ['concpm25']
        for var_name in var_names_to_test:
            r = ReadEEAAQEREP_V2()
            # r.logger.setLevel(logging.INFO)
            data = None
            data = r.read(vars_to_retrieve = [var_name])
            print('{} data read'.format(var_name))
            try:
                stat_data = data[station_id[var_name]]
                print('{} @ station {} mean: {} [{}]'.format(var_name, station_id[var_name],
                                                             stat_data[var_name].mean(),
                                                             stat_data['var_info'][var_name]['units']))
                # print(data[station_id[var_name]])
                # print(data[station_id[var_name]][var_name])
            except:
                print('failed test var {}'.format(var_name))
                pass

    elif username == 'jonasg':
        # Test that the reading routine works
        from pyaerocom.io.read_eea_aqerep_v2 import ReadEEAAQEREP_V2

        ddir = '/home/jonasg/MyPyaerocom/data/obsdata/EEA_AQeRep.NRT/download'
        reader = ReadEEAAQEREP_V2(data_dir=ddir)

        data = reader.read(['conco3'], last_file=1)




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
I/O helper methods of the pyaerocom package
"""
from collections import OrderedDict as od
from datetime import datetime
import os
from pathlib import Path
import shutil
from time import time

from pyaerocom import const
from pyaerocom.io import AerocomBrowser
from pyaerocom.exceptions import (VarNotAvailableError,
                                  VariableDefinitionError,
                                  )

#: country code file name
#: will be prepended with the path later on
COUNTRY_CODE_FILE = 'country_codes.json'

def _check_ebas_db_local_vs_remote(loc_remote, loc_local):
    """
    Check and if applicable, copy ebas_file_index.sqlite3 into cache dir

    Note
    ----
    This may speedup things if remote location is on a mounted server location.
    Nothing the user should worry about in any case.

    Parameters
    ----------
    loc_remote : str
        remote location of ebas_file_index.sqlite3
    loc_local : str
        local (cached) location of ebas_file_index.sqlite3

    Returns
    -------
    str
        valid location of ebas_file_index.sqlite3 that is supposed to be used

    """
    if os.path.exists(loc_remote): # remote exists
        if os.path.exists(loc_local):
            chtremote = os.path.getmtime(loc_remote)
            chtlocal = os.path.getmtime(loc_local)
            if chtlocal == chtremote:
                return loc_local

        # changing time differs -> try to copy to local and if that
        # fails, use remote location
        try:
            t0 = time()
            shutil.copy2(loc_remote, loc_local)
            const.print_log.info('Copied EBAS SQL database to {}\n'
                                'Elapsed time: {:.3f} s'
                                .format(loc_local, time() - t0))

            return loc_local
        except Exception as e:
            const.print_log.warning('Failed to copy EBAS SQL database. '
                                   'Reason: {}'.format(repr(e)))
            return loc_remote
    return loc_remote

def aerocom_savename(data_id, var_name, vert_code, year, ts_type):
    """Generate filename in AeroCom conventions

    ToDo: complete docstring
    """
    return ('aerocom3_{}_{}_{}_{}_{}.nc'
            .format(data_id, var_name, vert_code, year, ts_type))

def _print_read_info(i, mod, tot_num, last_t, name, logger):
    """Helper for displaying standardised output in reading classes

    Not to be used directly
    """
    t = datetime.now()
    logger.info("Reading files {}-{} of {} ({}) | {} (delta = {} s')"
                .format(i+1,i+1+mod, tot_num, name,
                        t.strftime('%H:%M:%S'),
                        (t-last_t).seconds))
    return t

def get_metadata_from_filename(filename):
    """Try access metadata information from filename"""
    from pyaerocom.io.fileconventions import FileConventionRead
    fc = FileConventionRead().from_file(filename)
    return fc.get_info_from_file(filename)

def read_ebas_flags_file(ebas_flags_csv):
    """Reads file ebas_flags.csv

    Parameters
    ----------
    ebas_flags_csv : str
        file containing flag info

    Returns
    -------
    dict
        dict with loaded flag info
    """
    valid = {}
    values = {}
    info = {}
    with open(ebas_flags_csv) as fio:
        for line in fio:
            spl = line.strip().split(',')
            num = int(spl[0].strip())
            try:
                val_str = spl[-1][1:-1]
            except Exception:
                raise IOError('Failed to read flag information in row {} '
                              '(Check if entries in ebas_flags.csv are quoted)'
                              .format(line))
            info_str = ','.join(spl[1:-1])
            try:
                info_str = info_str[1:-1]
            except Exception:
                raise IOError('Failed to read flag information in row {} '
                              '(Check if entries in ebas_flags.csv are quoted)'
                              .format(line))
            isvalid = True if val_str == 'V' else False
            valid[num] = isvalid
            values[num] = val_str
            info[num] = info_str
    result = {}
    result['valid'] = valid
    result['info'] = info
    result['vals'] = values
    return result

def add_file_to_log(filepath, err_msg):
    """
    Add input file path to error logdir

    The logdir location can be accessed via :attr:`pyaerocom.const.LOGFILESDIR`

    Parameters
    ----------
    filepath : str or Path
        path of file that has an error
    err_msg : str
        Problem associated with input file

    """
    if isinstance(filepath, Path):
        filepath = str(filepath)
    try:
        dirname = os.path.dirname(filepath)
        spl = dirname.split(os.sep)
        if spl[-1].lower() == 'renamed':
            model_or_obs_id = spl[-2]
        else:
            model_or_obs_id = spl[-1]
    except Exception:
        model_or_obs_id = 'others'

    logdir = const.LOGFILESDIR

    logfile = os.path.join(logdir, f'{model_or_obs_id}.log')

    if os.path.exists(logfile): #check if this file is already flagged
        with open(logfile, 'r') as f:
            for line in f:
                if filepath == line.strip():
                    return #file is already flagged -> ignore

    logfile_err = os.path.join(logdir, f'{model_or_obs_id}_ERR.log')
    with open(logfile, 'a+') as f:
        f.write(f'{filepath}\n')
    with open(logfile_err, 'a+') as ferr:
        ferr.write(f'{filepath}\n{err_msg}\n\n')

def get_standard_name(var_name):
    """Get standard name of aerocom variable

    Parameters
    ----------
    var_name : str
        HTAP2 variable name

    Returns
    --------
    str
        corresponding standard name

    Raises
    ------
    VarNotAvailableError
        if input variable is not defined in *variables.ini* file
    VariableDefinitionError
        if standarad name is not set for variable in *variables.ini* file
    """
    if not var_name in const.VARS:
        raise VarNotAvailableError('No such variable {}. Check variables.ini'.format(var_name))
    name = const.VARS[var_name].standard_name
    if name is None:
        raise VariableDefinitionError('standard_name not defined for variable')
    return name

def search_data_dir_aerocom(name_or_pattern, ignorecase=True):
    """Search Aerocom data directory based on model / data ID
    """
    browser = AerocomBrowser()
    return browser.find_data_dir(name_or_pattern, ignorecase)

def get_all_supported_ids_ungridded():
    """Get list of datasets that are supported by :class:`ReadUngridded`

    Returns
    -------
    list
        list with supported network names
    """
    from pyaerocom.io import ReadUngridded
    return ReadUngridded().SUPPORTED_DATASETS

def get_obsnetwork_dir(obs_id):
    """Returns data path for obsnetwork ID

    Parameters
    ----------
    obs_id : str
        ID  of obsnetwork (e.g. AeronetSunV2Lev2.daily)

    Returns
    -------
    str
        corresponding directory from ``pyaerocom.const``

    Raises
    ------
    ValueError
        if obs_id is invalid
    IOError
        if directory does not exist
    """
    if not obs_id in const.OBSLOCS_UNGRIDDED:
        raise ValueError(f"Observation network ID {obs_id} does not exist")

    data_dir = const.OBSLOCS_UNGRIDDED[obs_id]
    if not os.path.exists(data_dir):
        raise FileNotFoundError(
            f'Data directory {data_dir} for observation network {obs_id} does '
            f'not exist')
    return data_dir

def search_names(update_inifile=True, check_nc_file=True):
    """Search model IDs in database

    Parameters
    ----------
    update_inifile : bool
        if True, the file *names.txt* will be updated. The file is located
        in the installation *data* directory.
    check_nc_file : bool
        If True, only model IDs are included, for which at least one nc file
        can be detected in the corresponding renamed sub directory
    """
    names = []
    for mdir in const.DATA_SEARCH_DIRS:
        print("\n%s\n" %mdir)
        sub = os.listdir(mdir)
        for item in sub:
            path = os.path.join(mdir, item, "renamed")
            if os.path.isdir(path):
                print("\n%s\n" %path)
                add = True
                if check_nc_file:
                    add = False
                    for name in os.listdir(path):
                        if name.endswith(".nc"):
                            add = True
                            break
                if add:
                    names.append(item)
    names = sorted(od.fromkeys(names))
    if update_inifile:
        from pyaerocom import __dir__
        fpath = os.path.join(__dir__, "data", "names.txt")
        f = open(fpath, "w")
        for name in names:
            f.write("%s\n" %name)
        f.close()
    return names

def get_all_names():
    """Try to import all model IDs from file names.txt in data directory"""
    from pyaerocom import __dir__
    try:
        with open(os.path.join(__dir__, "data", "names.txt")) as f:
            names = f.read().splitlines()
        f.close()
    except Exception:
        try:
            names = search_names()
        except Exception:
            raise Exception("Failed to access model IDs")
    return names

def get_country_name_from_iso(iso_code=None,
                              filename=None,
                              return_as_dict=False):
    """get the country name from the 2 digit iso country code

    the underlaying json file was taken from this github repository
    https://github.com/lukes/ISO-3166-Countries-with-Regional-Codes

    Parameters
    ----------
    iso_code : :obj:`str`
        string containing the 2 character iso code of the country (e.g. no for Norway)
    filename : :obj:`str` , optional
        optional string with the json file to read
    return_as_dict : :obj:`bool`, optional
        flag to get the entire list of countries as a dictionary with the country codes
        as keys and the country names as value
        Useful if you have to get the names for a lot of country codes

    Returns
    -------
    string with country name or dictionary with iso codes as keys and the country names as values
    empty string if the country code was not found


    Raises
    ------
    ValueError
        if the country code ins invalid
    """
    if iso_code is None:
        return_as_dict = True

    if filename is None:
        #set default file name
        from pyaerocom import __dir__
        filename = os.path.join(__dir__, 'data', COUNTRY_CODE_FILE)

    import simplejson as json
    with open(filename) as fh:
        json_data = json.load(fh)

    iso_dict = {}
    for indict in json_data:
        iso_dict[indict['alpha-2']] = indict['name']

    if return_as_dict:
        return iso_dict
    else:
        try:
            ret_val = iso_dict[iso_code.upper()]
        except KeyError:
            ret_val = ''
            raise ValueError
        return ret_val



if __name__=="__main__":
    #names = search_names()
    names = get_all_names()



"""
Metadata keys from GHOST that only depend on station dimension

THESE SHOULD BE ALWAYS THE SAME ACCORDING TO INFORMATION INFORMATION
PROVIDED BY BSC
"""

GHOST_META_KEYS = [
    'EDGAR_v4.3.2_annual_average_BC_emissions',
    'EDGAR_v4.3.2_annual_average_CO_emissions',
    'EDGAR_v4.3.2_annual_average_NH3_emissions',
    'EDGAR_v4.3.2_annual_average_NMVOC_emissions',
    'EDGAR_v4.3.2_annual_average_NOx_emissions',
    'EDGAR_v4.3.2_annual_average_OC_emissions',
    'EDGAR_v4.3.2_annual_average_PM10_emissions',
    'EDGAR_v4.3.2_annual_average_SO2_emissions',
    'EDGAR_v4.3.2_annual_average_biogenic_PM2.5_emissions',
    'EDGAR_v4.3.2_annual_average_fossilfuel_PM2.5_emissions',
    'ESDAC_Iwahashi_landform_classification',
    'ESDAC_Meybeck_landform_classification',
    'ESDAC_modal_Iwahashi_landform_classification_25km',
    'ESDAC_modal_Iwahashi_landform_classification_5km',
    'ESDAC_modal_Meybeck_landform_classification_25km',
    'ESDAC_modal_Meybeck_landform_classification_5km',
    'ETOPO1_altitude',
    'ETOPO1_max_altitude_difference_5km',
    'GHOST_version',
    'GPW_average_population_density_25km',
    'GPW_average_population_density_5km',
    'GPW_max_population_density_25km',
    'GPW_max_population_density_5km',
    'GPW_population_density',
    'GSFC_coastline_proximity',
    'Joly-Peuch_classification_code',
    'Koppen-Geiger_classification',
    'Koppen-Geiger_modal_classification_25km',
    'Koppen-Geiger_modal_classification_5km',
    'MODIS_MCD12C1_v6_IGBP_land_use',
    'MODIS_MCD12C1_v6_LAI',
    'MODIS_MCD12C1_v6_UMD_land_use',
    'MODIS_MCD12C1_v6_modal_IGBP_land_use_25km',
    'MODIS_MCD12C1_v6_modal_IGBP_land_use_5km',
    'MODIS_MCD12C1_v6_modal_LAI_25km',
    'MODIS_MCD12C1_v6_modal_LAI_5km',
    'MODIS_MCD12C1_v6_modal_UMD_land_use_25km',
    'MODIS_MCD12C1_v6_modal_UMD_land_use_5km',
    'NOAA-DMSP-OLS_v4_average_nighttime_stable_lights_25km',
    'NOAA-DMSP-OLS_v4_average_nighttime_stable_lights_5km',
    'NOAA-DMSP-OLS_v4_max_nighttime_stable_lights_25km',
    'NOAA-DMSP-OLS_v4_max_nighttime_stable_lights_5km',
    'NOAA-DMSP-OLS_v4_nighttime_stable_lights',
    'OMI_level3_column_annual_average_NO2',
    'OMI_level3_column_cloud_screened_annual_average_NO2',
    'OMI_level3_tropospheric_column_annual_average_NO2',
    'OMI_level3_tropospheric_column_cloud_screened_annual_average_NO2',
    'UMBC_anthrome_classification',
    'UMBC_modal_anthrome_classification_25km',
    'UMBC_modal_anthrome_classification_5km',
    'WMO_region',
    'WWF_TEOW_biogeographical_realm',
    'WWF_TEOW_biome',
    'WWF_TEOW_terrestrial_ecoregion',
    'administrative_country_division_1',
    'administrative_country_division_2',
    'altitude',
    'area_classification',
    'associated_networks',
    'city',
    'climatology',
    'contact_email_address',
    'contact_institution',
    'contact_name',
    'country',
    'daily_passing_vehicles',
    'data_level',
    'daytime_traffic_speed',
    'distance_to_building',
    'distance_to_junction',
    'distance_to_kerb',
    'distance_to_source',
    'land_use',
    'latitude',
    'longitude',
    'main_emission_source',
    'measurement_altitude',
    'measurement_methodology',
    'measurement_scale',
    'measuring_instrument_calibration_scale',
    'measuring_instrument_documented_absorption_cross_section',
    'measuring_instrument_documented_accuracy',
    'measuring_instrument_documented_flow_rate',
    'measuring_instrument_documented_lower_limit_of_detection',
    'measuring_instrument_documented_measurement_resolution',
    'measuring_instrument_documented_precision',
    'measuring_instrument_documented_span_drift',
    'measuring_instrument_documented_uncertainty',
    'measuring_instrument_documented_upper_limit_of_detection',
    'measuring_instrument_documented_zero_drift',
    'measuring_instrument_documented_zonal_drift',
    'measuring_instrument_further_details',
    'measuring_instrument_inlet_information',
    'measuring_instrument_manual_name',
    'measuring_instrument_name',
    'measuring_instrument_process_details',
    'measuring_instrument_reported_absorption_cross_section',
    'measuring_instrument_reported_accuracy',
    'measuring_instrument_reported_flow_rate',
    'measuring_instrument_reported_lower_limit_of_detection',
    'measuring_instrument_reported_measurement_resolution',
    'measuring_instrument_reported_precision',
    'measuring_instrument_reported_span_drift',
    'measuring_instrument_reported_uncertainty',
    'measuring_instrument_reported_units',
    'measuring_instrument_reported_upper_limit_of_detection',
    'measuring_instrument_reported_zero_drift',
    'measuring_instrument_reported_zonal_drift',
    'measuring_instrument_sampling_type',
    'network',
    'network_maintenance_details',
    'network_miscellaneous_details',
    'network_provided_volume_standard_pressure',
    'network_provided_volume_standard_temperature',
    'network_qa_details',
    'network_sampling_details',
    'network_uncertainty_details',
    'population',
    'primary_sampling_further_details',
    'primary_sampling_instrument_documented_flow_rate',
    'primary_sampling_instrument_manual_name',
    'primary_sampling_instrument_name',
    'primary_sampling_instrument_reported_flow_rate',
    'primary_sampling_process_details',
    'primary_sampling_type',
    'principal_investigator_email_address',
    'principal_investigator_institution',
    'principal_investigator_name',
    'process_warnings',
    'representative_radius',
    'sample_preparation_further_details',
    'sample_preparation_process_details',
    'sample_preparation_techniques',
    'sample_preparation_types',
    'sampling_height',
    'station_classification',
    'station_name',
    'station_reference',
    'station_timezone',
    'street_type',
    'street_width',
    'terrain',
]



import xarray as xr
import numpy as np
import os
import glob

from pyaerocom import const
from pyaerocom.exceptions import VarNotAvailableError
from pyaerocom.io._read_mscw_ctm_helpers import (add_dataarrays,
                                                 subtract_dataarrays,
                                                 calc_concNhno3,
                                                 calc_concNno3pm10,
                                                 calc_concNno3pm25,
                                                 calc_conNtno3, calc_concNnh3,
                                                 calc_concNnh4, calc_concNtnh,
                                                 update_EC_units,
                                                 calc_concsspm25, calc_vmrox
                                                 )
from pyaerocom.variable_helpers import get_emep_variables
from pyaerocom.griddeddata import GriddedData
from pyaerocom.units_helpers import UALIASES


class ReadMscwCtm(object):
    """
    Class for reading model output from the EMEP MSC-W chemical transport model.

    Parameters
    ----------
    filepath : str
        Path to netcdf file.
    data_id : str
        string ID of model (e.g. "AATSR_SU_v4.3","CAM5.3-Oslo_CTRL2016")
    data_dir : str, optional
        Base directory of EMEP data, containing one or more netcdf files

    Attributes
    ----------
    data_id : str
        ID of model
    filename : str
        name of data file to be read.
    """
    # dictionary containing information about additionally required variables
    # for each auxiliary variable (i.e. each variable that is not provided
    # by the original data but computed on import)
    AUX_REQUIRES = {'depso4' : ['dryso4','wetso4'],
                    'concbc' : ['concbcf', 'concbcc'],
                    'concno3' : ['concno3c', 'concno3f'],
                    'concoa' : ['concoac', 'concoaf'],
                    'concpmgt25': ['concpm10', 'concpm25'],
                    'concNhno3' : ['conchno3'],
                    'concNnh3'  : ['concnh3'],
                    'concNnh4'  : ['concnh4'],
                    'concNno3pm10' : ['concno3f','concno3c'],
                    'concNno3pm25' : ['concno3f','concno3c'],
                    'concNtno3'   : ['conchno3','concno3f','concno3c'],
                    'concNtnh'    : ['concnh3','concnh4'],
                    'concsspm25'  : ['concssf', 'concssc'],
                    'concsspm10'  : ['concsspm25','concssc'],
                    'concCecpm25' : ['concecpm25'],
                    'vmrox'       : ['concno2', 'vmro3'],
                    }

    # Functions that are used to compute additional variables (i.e. one
    # for each variable defined in AUX_REQUIRES)
    # NOTE: these methods are supposed to work for xarray.DataArray instances
    # not iris.cube.Cube instance
    AUX_FUNS = {'depso4' : add_dataarrays,
                'concbc' : add_dataarrays,
                'concno3' : add_dataarrays,
                'concoa' : add_dataarrays,
                'concpmgt25': subtract_dataarrays,
                'concNhno3': calc_concNhno3,
                'concNnh3' : calc_concNnh3,
                'concNnh4' : calc_concNnh4,
                'concNno3pm10' : calc_concNno3pm10,
                'concNno3pm25' : calc_concNno3pm25,
                'concNtno3'    : calc_conNtno3,
                'concNtnh'     : calc_concNtnh,
                'concsspm25'   : calc_concsspm25,
                'concsspm10'   : add_dataarrays,
                'concCecpm25'  : update_EC_units,
                'vmrox'        : calc_vmrox,
                }

    #: supported filename masks, placeholder is for frequencies
    FILE_MASKS = ['Base_*.nc']

    #: frequencies encoded in filenames
    FREQ_CODES = {
        'hour'    : 'hourly',
        'day'     : 'daily',
        'month'   : 'monthly',
        'fullrun' : 'yearly',

    }

    DEFAULT_FILE_NAME = 'Base_day.nc'

    def __init__(self, filepath=None, data_id=None, data_dir=None):
        self._data_dir = None
        # opened dataset (for performance boost), will be reset if data_dir is
        # changed
        self._filename = None
        self._filedata = None

        self._file_mask = None
        self._files = None

        self.var_map = get_emep_variables()

        data_dir, filename, data_id = self._eval_input(filepath, data_id,
                                                       data_dir)
        self.data_id = data_id
        if data_dir is not None:
            self.data_dir = data_dir

        if filename is None:
            filename = self.DEFAULT_FILE_NAME

        self.filename = filename

    def _eval_input(self, filepath, data_id, data_dir):
        """
        Evaluate input (helper method for __init__)

        Note, this method does not change the associated class attributes, it
        just does some sanity checking on what the user inputs.

        Parameters
        ----------
        filepath : str, optional
            path to file to be read
        data_id : str, optional
            ID of dataset
        data_dir : str, optional
            directory containing EMEP data files

        Raises
        ------
        FileNotFoundError
            if any of input data_dir or filepath are provided but do not exist
        ValueError
            if any of input data_dir or filepath are provided but are not
            a directory or file, respectively.

        Returns
        -------
        tuple
            3-element tuple containing (potentially updated / inferred values of):

                - `data_dir`
                - `filename`
                - `data_id`

        """
        filename = None
        if filepath is not None:
            if not isinstance(filepath, str) or not os.path.exists(filepath):
                raise FileNotFoundError(f'{filepath}')
            if os.path.isdir(filepath):
                raise ValueError(f'{filepath} is a directory, please use data_dir')
            data_dir,filename = os.path.split(filepath)

        if data_dir is not None:
            if not isinstance(data_dir, str) or not os.path.exists(data_dir):
                raise FileNotFoundError(f'{data_dir}')
            if not os.path.isdir(data_dir):
                raise ValueError(f'{data_dir} is not a directory')
        if data_id is None and data_dir is not None:
            data_id = data_dir.split(os.sep)[-1]
        return (data_dir,filename,data_id)

    @property
    def data_dir(self):
        """
        Directory containing netcdf files
        """
        return self._data_dir

    @data_dir.setter
    def data_dir(self, val):
        if not os.path.isdir(val):
            raise FileNotFoundError(val)
        mask, filelist = self._check_files_in_data_dir(val)
        self._file_mask = mask
        self._files = filelist
        self._data_dir = val
        self._filedata = None

    @property
    def filename(self):
        """
        Name of netcdf file
        """
        return self._filename

    @filename.setter
    def filename(self, val):
        """
        Name of netcdf file
        """
        if not isinstance(val, str):
            raise ValueError('need str')
        elif val == self._filename:
            return
        self._filename = val
        self._filedata = None

    @property
    def filepath(self):
        """
        Path to data file
        """
        if self.data_dir is None:
            raise AttributeError('need data_dir to be set in data')
        return os.path.join(self.data_dir, self.filename)

    @filepath.setter
    def filepath(self, value):
        ddir, fname = os.path.split(value)
        self.data_dir = ddir
        self.filename = fname

    @property
    def filedata(self):
        """
        Loaded netcdf file (:class:`xarray.Dataset`)
        """
        if self._filedata is None:
            self.open_file()
        return self._filedata

    def _check_files_in_data_dir(self, data_dir):
        """
        Check for data files in input data directory

        Parameters
        ----------
        data_dir : str
            directory to be searched.

        Raises
        ------
        FileNotFoundError
            if no EMEP files can be identified

        Returns
        -------
        str
            file mask used (is identified automatically based on
            :attr:`FILE_MASKS`)
        list
            list of file matches

        """

        for fmask in self.FILE_MASKS:
            matches = glob.glob(f'{data_dir}/{fmask}')
            if len(matches) > 0:
                return fmask, matches
        raise FileNotFoundError(
            f'No valid model files could be found in {data_dir} for any of the '
            f'supported file masks: {self.FILE_MASKS}')

    @property
    def ts_type(self):
        """
        Frequency of time dimension of current data file

        Raises
        ------
        AttributeError
            if :attr:`filename` is not set.

        Returns
        -------
        str
            current ts_type.

        """
        return self.ts_type_from_filename(self.filename)

    @property
    def ts_types(self):
        """
        List of available frequencies

        Raises
        ------
        AttributeError
            if :attr:`data_dir` is not set.

        Returns
        -------
        list
            list of available frequencies

        """
        if not isinstance(self._files, list):
            raise AttributeError('please set data_dir first')
        tsts = []
        for file in self._files:
            tsts.append(self.ts_type_from_filename(file))
        return tsts

    @property
    def years_avail(self):
        """
        Years available in loaded dataset
        """
        data = self.filedata
        years = data.time.dt.year.values
        years = list(np.unique(years))
        return sorted(years)

    @property
    def vars_provided(self):
        """Variables provided by this dataset"""
        return list(self.var_map) + list(self.AUX_REQUIRES)

    def open_file(self):
        """
        Open current netcdf file

        Returns
        -------
        xarray.Dataset

        """
        fp = self.filepath
        const.print_log.info(f'Opening {fp}')

        ds = xr.open_dataset(fp)
        self._filedata = ds
        return ds

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return 'ReadMscwCtm'


    def has_var(self, var_name):
        """Check if variable is supported

        Parameters
        ----------
        var_name : str
            variable to be checked

        Returns
        -------
        bool
        """
        avail = self.vars_provided
        if var_name in avail or const.VARS[var_name].var_name_aerocom in avail:
            return True
        return False

    def ts_type_from_filename(self, filename):
        """
        Get ts_type from filename

        Parameters
        ----------
        filename : str

        Raises
        ------
        ValueError
            if ts_type cannot be inferred from filename.

        Returns
        -------
        tstype : str
        """
        filename = os.path.basename(filename)
        for substr, tstype in self.FREQ_CODES.items():
            if substr in filename:
                return tstype
        raise ValueError(f'Failed to retrieve ts_type from filename {filename}')

    def filename_from_ts_type(self, ts_type):
        """
        Infer file name of data based on input ts_type

        Parameters
        ----------
        ts_type : str
            desired time freq of data

        Raises
        ------
        ValueError
            If no file could be inferred.

        Returns
        -------
        fname : str
            Name of data file inferred based on current file mask and input
            freq.

        """
        mask = self._file_mask
        for substr, tst in self.FREQ_CODES.items():
            if tst == ts_type:
                fname = mask.replace('*', substr)
                return fname
        raise ValueError('failed to infer filename from input ts_type={ts_type}')

    def _compute_var(self, var_name_aerocom, ts_type):
        """Compute auxiliary variable

        Like :func:`read_var` but for auxiliary variables
        (cf. AUX_REQUIRES)

        Parameters
        ----------
        var_name : str
            variable that are supposed to be read
        ts_type : str
            string specifying temporal resolution.

        Returns
        -------
        GriddedData
            loaded data object
        """

        temp_arrs = []
        req = self.AUX_REQUIRES[var_name_aerocom]
        aux_func = self.AUX_FUNS[var_name_aerocom]
        const.print_log.info(
                f'computing {var_name_aerocom} from {req} using {aux_func}'
                )
        for aux_var in self.AUX_REQUIRES[var_name_aerocom]:
            arr = self._load_var(aux_var, ts_type)
            temp_arrs.append(arr)

        return aux_func(*temp_arrs)

    def _load_var(self, var_name_aerocom, ts_type):
        """
        Load variable data as :class:`xarray.DataArray`.

        This combines both, variables that can be read directly and auxiliary
        variables that are computed.

        Parameters
        ----------
        var_name_aerocom : str
            variable name
        ts_type : str
            desired frequency

        Raises
        ------
        VarNotAvailableError
            if input variable is not available

        Returns
        -------
        xarray.DataArray
            loaded data

        """
        if var_name_aerocom in self.var_map: #can be read
            return self._read_var_from_file(var_name_aerocom, ts_type)
        elif var_name_aerocom in self.AUX_REQUIRES:
            return self._compute_var(var_name_aerocom, ts_type)
        raise VarNotAvailableError('Variable {var_name} is not supported')

    def read_var(self, var_name, ts_type=None, **kwargs):
        """Load data for given variable.

        Parameters
        ----------
        var_name : str
            Variable to be read
        ts_type : str
            Temporal resolution of data to read. Supported are
            "hourly", "daily", "monthly" , "yearly".

        Returns
        -------
        GriddedData
        """
        if not self.has_var(var_name):
            raise VarNotAvailableError(var_name)
        var = const.VARS[var_name]
        var_name_aerocom = var.var_name_aerocom

        if self.data_dir is None:
            raise ValueError('data_dir must be set before reading.')
        elif self.filename is None and ts_type is None:
            raise ValueError('please specify ts_type')
        elif ts_type is not None:
            #filename and ts_type are set. update filename if ts_type suggests
            #that current file has different resolution
            self.filename = self.filename_from_ts_type(ts_type)

        ts_type = self.ts_type

        arr = self._load_var(var_name_aerocom, ts_type)
        if arr.units in UALIASES:
            arr.attrs['units'] = UALIASES[arr.units]
        try:
            cube = arr.to_iris()
        except MemoryError as e:
            raise NotImplementedError(f'BAAAM: {e}')

        if ts_type == 'hourly':
            cube.coord('time').convert_units('hours since 1900-01-01')
        gridded = GriddedData(cube, var_name=var_name_aerocom,
                              ts_type=ts_type, check_unit=True,
                              convert_unit_on_init=True)

        #!obsolete
        #if var.is_deposition:
        #    implicit_to_explicit_rates(gridded, ts_type)

        # At this point a GriddedData object with name gridded should exist

        gridded.metadata['data_id'] = self.data_id
        gridded.metadata['from_files'] = [self.filepath]

        # Remove unneccessary metadata. Better way to do this?
        for metadata in ['current_date_first', 'current_date_last']:
            if metadata in gridded.metadata.keys():
                del(gridded.metadata[metadata])
        return gridded

    def _read_var_from_file(self, var_name_aerocom, ts_type):
        """
        Read variable data from file as :class:`xarray.DataArray`.

        See also :func:`_load_var`

        Parameters
        ----------
        var_name_aerocom : str
            variable name
        ts_type : str
            desired frequency

        Raises
        ------
        VarNotAvailableError
            if input variable is not available

        Returns
        -------
        xarray.DataArray
            loaded data

        """
        emep_var = self.var_map[var_name_aerocom]
        try:
            data = self.filedata[emep_var]
        except KeyError:
            raise VarNotAvailableError(
                f'{var_name_aerocom} ({emep_var}) not available in {self.filename}')
        data.attrs['long_name'] = var_name_aerocom
        data.time.attrs['long_name'] = 'time'
        data.time.attrs['standard_name'] = 'time'
        prefix = emep_var.split('_')[0]
        data.attrs['units'] = self.preprocess_units(data.units, prefix)
        return data

    @staticmethod
    def preprocess_units(units, prefix):
        """
        Update units for certain variables

        Parameters
        ----------
        units : str
            Current unit of data
        prefix : str, optional
            Variable prefix (e.g. AOD, AbsCoeff).

        Returns
        -------
        str
            updated unit (where applicable)

        """
        if units == '' and prefix == 'AOD': #
            return '1'
        elif units == '' and prefix == 'AbsCoef':
            return 'm-1'
        return units

class ReadEMEP(ReadMscwCtm):
    """Old name of :class:`ReadMscwCtm`."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        print("You are using a deprecated name ReadEMEP for class ReadMscwCtm")




if __name__ == '__main__': # pragma: no cover

    EMEP_DIR = '/lustre/storeB/project/fou/kl/emep/ModelRuns/2020_REPORTING/EMEP01_rv4_35_2018_emepCRef2_XtraOut/'

    fname = 'Base_month.nc'

    fp = EMEP_DIR + fname

    reader = ReadMscwCtm(data_dir=EMEP_DIR)#+'Base_month.nc')


    # Read variable that uses AUX_FUNS
    data = reader.read_var('concno3', ts_type='daily')



from setuptools import setup

with open('README.rst') as file:
    readme = file.read()

with open('VERSION.md') as f:
    version = f.readline()
    print(version)
    f.close()

setup(
    name        =   'pyaerocom',
    version     =   version,
    author      =   'MET Norway',
    url         =   'https://github.com/metno/pyaerocom.git',
    license     =   'GPLv3',
    package_dir =   {'pyaerocom'      :   'pyaerocom'},
    packages    =   ['pyaerocom',
                     'pyaerocom.io',
                     'pyaerocom.plot',
                     'pyaerocom.tools',
                     'pyaerocom.aeroval',
                     'pyaerocom.scripts'],
    package_data=   {'pyaerocom'    :   ['data/*']},

    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        # Indicate who your project is intended for
        'Intended Audience :: Science/Research',
        'Intended Audience :: Education',
        'Topic :: Scientific/Engineering :: Atmospheric Science',

        # Pick your license as you wish (should match 'license' above)
        'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.,
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],

    python_requires = '>=3.7,<4',
    install_requires = [
        'scitools-iris>=3.0.1,<3.1.0',
        'scitools-pyke>=1.1.1',
        'xarray>=0.16.0',
        'cartopy>=0.16.0',
        'matplotlib>=3.0.1',
        'scipy>=1.1.0',
        'pandas>=0.23.0',
        'seaborn>=0.8.0',
        'geonum',
        'LatLon23', # required by geonum
        'SRTM.py', # required by geonum
        'numpy',
        'simplejson',
        'requests',
        'reverse-geocode',
        'tqdm',
        'openpyxl',
        'geojsoncontour'
    ],
    extras_require = {
        'docs':['nbsphinx'],
        'test':['pytest>=6.0','pytest-dependency','pytest-cov'],
    },
    dependency_links    =   [],
    description = ('pyaerocom model evaluation software'),
    long_description = readme,
    long_description_content_type='text/x-rst',
    entry_points = {'console_scripts' : [
            'pya=pyaerocom.scripts.cli:main'
            ]},
    zip_safe = False
)



VAR_PREFIXES = ['vmr', 'mmr', 'conc', 'sconc', 'wet', 'dry']

# in g/mol
MOLMASSES = {'air_dry'  : 28.9647,
             'o3'       : 48,
             'so2'      : 64.066,
             'so4'      : 96.06,
             'no'       : 30.01,
             'no2'      : 46.0055,
             'hno3'     : 63.01,
             'nh3'      : 17.031,
             'co'       : 28.010,
             'isop'     : 68.12,
             'glyoxal'  : 58.036,
             'glyox'    : 58.036,
             'hcho'     : 30.026,}

class UnkownSpeciesError(ValueError):
    pass

def get_species(var_name):
    """
    Get species name from variable name

    Parameters
    ----------
    var_name : str
        pyaerocom variable name (cf. variables.ini)

    Raises
    ------
    UnkownSpeciesError
        if species cannot be inferred

    Returns
    -------
    str
        name of species

    """
    if var_name in MOLMASSES:
        return var_name
    for prefix in VAR_PREFIXES:
        if var_name.startswith(prefix):
            species = var_name.split(prefix)[-1]
            if species in MOLMASSES:
                return species
    raise UnkownSpeciesError('Could not infer atom / molecule/ species from '
                             'var_name {}'.format(var_name))

def get_molmass(var_name):
    """
    Get molar mass for input variable

    Parameters
    ----------
    var_name : str
        pyaerocom variable name (cf. variables.ini) or name of species

    Returns
    -------
    float
        molar mass of species in units of g/mol

    """
    return MOLMASSES[get_species(var_name)]

def get_mmr_to_vmr_fac(var_name):
    """
    Get conversion factor for MMR -> VMR conversion for input variable

    Note
    ----
    Assumes dry air molar mass

    Parameters
    ----------
    var_name : str
        Name of variable to be converted

    Returns
    -------
    float
        multiplication factor to convert MMR -> VMR

    """
    return get_molmass('air_dry') / get_molmass(var_name)



from warnings import warn

_cmap_lighttheme = "Blues"

DEFAULT_THEME = "light"
_COLOR_THEMES = dict(light = dict(name="light",
                                  cmap_map=_cmap_lighttheme,
                                  cmap_map_div='bwr_r',
                                  color_coastline ="k"),
                     dark = dict(name="dark",
                                 cmap_map="viridis",
                                 cmap_map_div='PuOr_r',
                                 color_coastline ="#e6e6e6"))

MAP_AXES_ASPECT = 1.5
FIGSIZE_DEFAULT = (16, 10)
# text positions for annotations in scatter plots
SCAT_ANNOT_XYPOS = [(.01, 0.95),
                    (0.01, 0.90),
                    (0.3, 0.90),
                    (0.01, 0.86),
                    (0.3, 0.86),
                    (0.01, 0.82),
                    (0.3, 0.82),
                    (0.01, 0.78),
                    (0.3, 0.78),
                    (0.8, 0.1),
                    (0.8, 0.06)]

class ColorTheme(object):
    """Pyaerocom class specifiying plotting color theme

    Attributes
    ----------
    name : str
        name of color theme (e.g. "light" or "dark")
    cmap_map : str
        name of colormap or colormap for map plotting
    color_coastline : str
        coastline color for map plotting
    cmap_map_div : str
        name of diverging colormap (used in map plots when plotted value range
        crosses 0)
    cmap_map_div_shifted : bool
        boolean specifying whether center of diverging colormaps for map plots
        is supposed to be shifted to 0

    Example
    -------
    Load default color theme
    >>> theme = ColorTheme(name="dark")
    >>> print(theme)
    pyaerocom ColorTheme
    name : dark
    cmap_map : viridis
    color_coastline : #e6e6e6
    """
    def __init__(self, name="dark", cmap_map=None, color_coastline=None,
                 cmap_map_div=None, cmap_map_div_shifted=True):
        if not name in _COLOR_THEMES:
            warn("Invalid name for color theme, using default theme")
            name = DEFAULT_THEME

        self.name = name
        self.cmap_map = cmap_map
        self.cmap_map_div = cmap_map_div
        self.cmap_map_div_shifted = cmap_map_div_shifted
        self.color_coastline = color_coastline
        self.color_map_text = 'r'

        self.load_default(name)

    def load_default(self, theme_name="dark"):
        """Load default color theme

        Parameters
        ----------
        theme_name : str
            name of default theme

        Raises
        ------
        ValueError
            if ``theme_name`` is not a valid default theme

        """
        if not theme_name in _COLOR_THEMES:
            raise ValueError("Default theme with name %s is not available. "
                             "Choose from %s" %_COLOR_THEMES)
        self.from_dict(_COLOR_THEMES[theme_name])

    def from_dict(self, info_dict):
        """Import theme information from dictionary

        info_dict : dict
            dictionary containing theme settings
        """
        for k, v in info_dict.items():
            if k in self.__dict__:
                self[k] = v

    def to_dict(self):
        """Convert this object into dictionary

        Returns
        -------
        dict
            dictionary representation of this object

        """
        d = {}
        for k, v in self.__dict__.items():
            d[k] = v
        return d

    def __setitem__(self, key, value):
        if key in self.__dict__:
            self.__dict__[key] = value

    def __repr__(self):
        return self.name

    def __str__(self):
        s = "pyaerocom ColorTheme"
        for k, v in self.__dict__.items():
            s += "\n%s : %s" %(k, v)
        return s

COLOR_THEME = ColorTheme(DEFAULT_THEME)

def get_color_theme(theme_name="dark"):
# Settings for colormap (use perceptually uniform colormaps)
    if not theme_name in _COLOR_THEMES:
        raise ValueError("Invalid input for theme_name, choose from %s"
                         %_COLOR_THEMES)
    return ColorTheme(theme_name)

if __name__=="__main__":
    print(ColorTheme("dark"))
    print(ColorTheme("light"))

    import doctest
    doctest.testmod()



#Copyright (C) 2017 met.no
#Norwegian Meteorological Institute
#Box 43 Blindern
#0313 OSLO
#NORWAY
#E-mail: jan.griesfeller@met.no, jonas.gliss@met.no
#This program is free software; you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation; either version 3 of the License, or
#(at your option) any later version.
#This program is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#GNU General Public License for more details.
#You should have received a copy of the GNU General Public License
#along with this program; if not, write to the Free Software
#Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#MA 02110-1301, USA

"""
Module containing pyaerocom custom exceptions
"""
class AeroValConfigError(ValueError):
    pass

class AeroValTrendsError(ValueError):
    pass

class CachingError(IOError):
    pass

class CacheWriteError(CachingError):
    pass

class CacheReadError(CachingError):
    pass

class ColocationError(ValueError):
    pass

class ColocationSetupError(ValueError):
    pass

class CoordinateError(ValueError):
    pass

class CoordinateNameError(CoordinateError):
    pass

class DataRetrievalError(IOError):
    pass

class DataCoverageError(ValueError):
    pass

class DataDimensionError(ValueError):
    pass

class DataIdError(ValueError):
    pass

class DataQueryError(ValueError):
    pass

class DataSourceError(ValueError):
    pass

class DataUnitError(ValueError):
    pass

class DeprecationError(AttributeError, ValueError):
    pass

class DimensionOrderError(DataDimensionError):
    pass

class DataExtractionError(ValueError):
    pass

class DataSearchError(IOError):
    pass

class EvalEntryNameError(KeyError):
    pass

class NasaAmesReadError(IOError):
    pass

class EbasFileError(ValueError):
    pass

class EEAv2FileError(ValueError):
    pass

class EntryNotAvailable(KeyError):
    pass

class InitialisationError(ValueError):
    pass

class FileConventionError(IOError):
    pass

class LongitudeConstraintError(ValueError):
    pass

class MetaDataError(AttributeError):
    pass

class NetworkNotSupported(NotImplementedError):
    pass

class NetworkNotImplemented(NotImplementedError):
    pass

class NetcdfError(IOError):
    pass

class NotInFileError(IOError):
    pass

class ResamplingError(ValueError):
    pass

class StationNotFoundError(AttributeError):
    pass

class TimeZoneError(AttributeError):
    pass

class TimeMatchError(AttributeError):
    pass

class TemporalResolutionError(ValueError):
    pass

class TemporalSamplingError(ValueError):
    pass

class UnitConversionError(ValueError):
    pass

class UnknownRegion(ValueError):
    pass

class UnresolvableTimeDefinitionError(DataDimensionError, NetcdfError):
    """Is raised if time definition in NetCDF file is wrong and cannot be corrected"""
    pass

class VarNotAvailableError(DataCoverageError):
    pass

class VariableDefinitionError(IOError):
    pass

class VariableNotFoundError(IOError):
    pass



################################################################
# read_aeronet_sdav2.py
#
# read Aeronet SDA V2 data
#
# this file is part of the pyaerocom package
#
#################################################################
# Created 20171026 by Jan Griesfeller for Met Norway
#
# Last changed: See git log
#################################################################

# Copyright (C) 2017 met.no
# Contact information:
# Norwegian Meteorological Institute
# Box 43 Blindern
# 0313 OSLO
# NORWAY
# E-mail: jan.griesfeller@met.no
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA 02110-1301, USA

import numpy as np
import pandas as pd
import re, os
from collections import OrderedDict as od
from pyaerocom import const as const
from pyaerocom.aux_var_helpers import calc_ang4487aer, calc_od550aer, calc_od550gt1aer, calc_od550lt1aer
from pyaerocom.stationdata import StationData
from pyaerocom.io.readaeronetbase import ReadAeronetBase

class ReadAeronetSdaV2(ReadAeronetBase):
    """Interface for reading Aeronet Sun V2 Level 2 data

    Note
    ----
    Maintenance for AERONET version 2 data is deprecated since pyaerocom
    version 0.11.0. Please use version 3 data and associated pyaerocom readers.
    """
    #: Mask for identifying datafiles
    _FILEMASK = '*.ONEILL_20'

    #: version log of this class (for caching)
    __version__ = '0.08_' + ReadAeronetBase.__baseversion__

    #: Name of dataset (OBS_ID)
    DATA_ID = const.AERONET_SUN_V2L2_SDA_DAILY_NAME

    #: List of all datasets supported by this interface
    SUPPORTED_DATASETS = [const.AERONET_SUN_V2L2_SDA_DAILY_NAME]

    #: dictionary assigning temporal resolution flags for supported datasets
    #: that are provided in a defined temporal resolution
    TS_TYPES = {const.AERONET_SUN_V2L2_SDA_DAILY_NAME   :  'daily'}

    #: default variables for read method
    DEFAULT_VARS = ['od550aer', 'od550gt1aer','od550lt1aer']

    #: value corresponding to invalid measurement
    NAN_VAL = -9999.

    # Comment jgliss 20180719: translated most varname to Aerocom convention.
    # Those, for which no defined name is known were named with a leading
    # underscore (e.g. _eta500lt1 for fine mode fraction at 500nm)
    #: Dictionary that specifies the index for each data column
    COL_INDEX = od(date                 = 0, # Date(dd:mm:yyyy)
                   time                 = 1, # Time(hh:mm:ss),
                   julien_day           = 2, # Julian_Day
                   od500aer             = 3, # Total_AOD_500nm[tau_a]
                   od500lt1aer          = 4, # Fine_Mode_AOD_500nm[tau_f]
                   od500gt1aer          = 5, # Coarse_Mode_AOD_500nm[tau_c]
                   _eta500lt1           = 6, # FineModeFraction_500nm[eta]
                   _aod500aer_fiterr    = 7, # 2nd_Order_Reg_Fit_Error-Total_AOD_500nm[regression_dtau_a]
                   _aod500lt1aer_rmse   = 8, # RMSE_Fine_Mode_AOD_500nm[Dtau_f]
                   _aod500gt1aer_rmse   = 9, # RMSE_Coarse_Mode_AOD_500nm[Dtau_c]
                   _eta500lt1_rmse      = 10, # RMSE_FineModeFraction_500nm[Deta]
                   _ang50aer            = 11, # Angstrom_Exponent(AE)-Total_500nm[alpha]
                   od870aer             = 15, # 870nm_Input_AOD
                   od675aer             = 16, # 675nm_Input_AOD
                   od667aer             = 17, # 667nm_Input_AOD
                   od555aer             = 18, # 555nm_Input_AOD
                   od551aer             = 19, # 551nm_Input_AOD
                   od532aer             = 20, # 532nm_Input_AOD
                   od531aer             = 21, # 531nm_Input_AOD
                   od500aer_input       = 22, # 500nm_Input_AOD
                   od490aer             = 23, # 490nm_Input_AOD
                   od443aer             = 24, # 443nm_Input_AOD
                   od440aer             = 25, # 440nm_Input_AOD
                   od412aer             = 26, # 412nm_Input_AOD
                   od380aer             = 27) # 380nm_Input_AOD

    # FURTHER HEADER COLUMN INDICES

    # i_dAE/dln(wavelength)-Total_500nm[alphap]=12,
    #i_AE-Fine_Mode_500nm[alpha_f]=13,
    #i_dAE/dln(wavelength)-Fine_Mode_500nm[alphap_f]=14,
    #LAST_PROCESSING_DATE_INDEX = 28, #Last_Processing_Date
    #NUMBER_OF_WAVELENGTHS_INDEX = 29, # Number_of_Wavelengths
    #UMBER_OF_OBSERVATIONS_INDEX = 30, #Number_of_Observations
    #EXACT_WAVELENGTHS_FOR_INPUT_AOD_INDEX = 31) # Exact_Wavelengths_for_Input_AOD

    # specify required dependencies for auxiliary variables, i.e. variables
    # that are NOT in Aeronet files but are computed within this class.
    # For instance, the computation of the AOD at 550nm requires import of
    # the AODs at 440, 500 and 870 nm.

    #: dictionary containing information about additionally required variables
    #: for each auxiliary variable (i.e. each variable that is not provided
    #: by the original data but computed on import)
    AUX_REQUIRES = {'ang4487aer'    :   ['od440aer',
                                         'od870aer'],
                    'od550aer'      :   ['od500aer', 'ang4487aer'],
                    'od550gt1aer'   :   ['od500gt1aer', 'ang4487aer'],
                    'od550lt1aer'   :   ['od500lt1aer', 'ang4487aer']}

    #: Functions that are used to compute additional variables (i.e. one
    #: for each variable defined in AUX_REQUIRES)
    AUX_FUNS = {'ang4487aer'    :   calc_ang4487aer,
                'od550aer'      :   calc_od550aer,
                'od550gt1aer'   :   calc_od550gt1aer,
                'od550lt1aer'   :   calc_od550lt1aer}

    #: List of variables that are provided by this dataset (will be extended
    #: by auxiliary variables on class init, for details see __init__ method of
    #: base class ReadUngriddedBase)
    PROVIDES_VARIABLES = list(COL_INDEX.keys())

    # This how the beginning of data file looks like

    # Level 2.0 Quality Assured Data. The following AERONET-SDA data are
    # derived from AOD data which are pre and post-field calibrated and
    # manually inspected. SDA Version 4.1 (tauf_tauc),Note: the labels in
    # square brackets that follow some of the parameter (column) names are
    # the symbols associated with these parameters in the original SDA
    # publication of O'Neill et al. (2003)
    # Location=Zvenigorod,Latitude=55.695000,Longitude=36.775000,
    # Elevation[m]=200.000000,PI=Brent_Holben,Email=Brent.N.Holben@nasa.gov
    # SDA from Level 2.0 AOD,Daily Average,UNITS can be found at
    # http://aeronet.gsfc.nasa.gov/data_menu.html
    # Date(dd:mm:yyyy),
    # Time(hh:mm:ss),
    # Julian_Day,Total_AOD_500nm[tau_a],Fine_Mode_AOD_500nm[tau_f],
    # Coarse_Mode_AOD_500nm[tau_c],FineModeFraction_500nm[eta],
    # 2nd_Order_Reg_Fit_Error-Total_AOD_500nm[regression_dtau_a],
    # RMSE_Fine_Mode_AOD_500nm[Dtau_f],RMSE_Coarse_Mode_AOD_500nm[Dtau_c],
    # RMSE_FineModeFraction_500nm[Deta],Angstrom_Exponent(AE)-
    # Total_500nm[alpha],dAE/dln(wavelength)-Total_500nm[alphap],
    # AE-Fine_Mode_500nm[alpha_f],
    # dAE/dln(wavelength)-Fine_Mode_500nm[alphap_f],
    # 870nm_Input_AOD,675nm_Input_AOD,667nm_Input_AOD,555nm_Input_AOD,
    # 551nm_Input_AOD,532nm_Input_AOD,531nm_Input_AOD,
    # 500nm_Input_AOD,490nm_Input_AOD,443nm_Input_AOD,440nm_Input_AOD,
    # 412nm_Input_AOD,380nm_Input_AOD,Last_Processing_Date,
    # Number_of_Wavelengths,Number_of_Observations,
    # Exact_Wavelengths_for_Input_AOD
    # 16:09:2006,00:00:00,259.000000,0.059239,0.032410,0.026828,0.571506,
    # 0.004645,0.007370,0.005064,0.091573,1.278830,-0.656569,2.354811,
    # 1.413517,0.036734,0.039337,-999.,-999.,-999.,-999.,-999.,0.064670,
    # -999.,-999.,0.069614,-999.,0.083549,27:11:2007,5,11,0.868800,
    # 0.675600,0.440400,0.500500,0.380100

    @property
    def col_index(self):
        """Dictionary that specifies the index for each data column

        Note
        ----
        Pointer to :attr:`COL_INDEX`
        """
        return self.COL_INDEX

    def read_file(self, filename, vars_to_retrieve=None,
                  vars_as_series=False):
        """Read Aeronet Sun SDA V2 file

        Parameters
        ----------
        filename : str
            absolute path to filename to read
        vars_to_retrieve : :obj:`list`, optional
            list of str with variable names to read. If None, use
            :attr:`DEFAULT_VARS`
        vars_as_series : bool
            if True, the data columns of all variables in the result dictionary
            are converted into pandas Series objects

        Returns
        -------
        StationData
            dict-like object containing results
        """
        if vars_to_retrieve is None:
            vars_to_retrieve = self.DEFAULT_VARS
        # implemented in base class
        vars_to_read, vars_to_compute = self.check_vars_to_retrieve(vars_to_retrieve)

        #create empty data object (is dictionary with extended functionality)
        data_out = StationData()

        data_out.data_id = self.data_id
        # create empty arrays for meta information
        for item in self.META_NAMES_FILE:
            data_out[item] = []

        # create empty arrays for all variables that are supposed to be read
        # from file
        for var in vars_to_read:
            data_out[var] = []

        # Iterate over the lines of the file
        self.logger.info("Reading file {}".format(filename))
        # Iterate over the lines of the file
        with open(filename, 'rt') as in_file:

            c_head_line = in_file.readline()
            c_algorithm = in_file.readline()

            self.logger.info(c_head_line)
            self.logger.info(c_algorithm)

            c_dummy = in_file.readline()
            # re.split(r'=|\,',c_dummy)
            i_dummy = iter(re.split(r'=|\,', c_dummy.rstrip()))
            dict_loc = dict(zip(i_dummy, i_dummy))

            data_out['latitude'] = float(dict_loc['Latitude'])
            data_out['longitude'] = float(dict_loc['Longitude'])
            data_out['altitude'] = float(dict_loc['Elevation[m]'])
            data_out['station_name'] = dict_loc['Location']
            data_out['PI'] = dict_loc['PI']
            data_out['ts_type'] = self.TS_TYPE

            c_dummy = in_file.readline()
            c_Header = in_file.readline()

            self.logger.info(c_Header)

            col_index = self.col_index

            vars_available = {}
            for var in vars_to_read:
                if var in col_index:
                    vars_available[var] = col_index[var]
                else:
                    self.logger.warning("Variable {} not available in file {}"
                                        .format(var, os.path.basename(filename)))

            for line in in_file:
                # process line
                dummy_arr = line.split(self.COL_DELIM)

                # This uses the numpy datestring64 functions that e.g. also
                # support Months as a time step for timedelta
                # Build a proper ISO 8601 UTC date string
                day, month, year = dummy_arr[col_index['date']].split(':')
                # pdb.set_trace()
                datestring = '-'.join([year, month, day])
                datestring = 'T'.join([datestring, dummy_arr[col_index['time']]])
                # NOTE JGLISS: parsing timezone offset was removed on 22/2/19
                # since it is deprecated in recent numpy versions, for details
                # see https://www.numpy.org/devdocs/reference/arrays.datetime.html#changes-with-numpy-1-11
                #datestring = '+'.join([datestring, '00:00'])
                data_out['dtime'].append(np.datetime64(datestring))

                # copy the data fields
                for var, idx in vars_available.items():
                    val = np.float_(dummy_arr[idx])
                    if val == self.NAN_VAL:
                        val = np.nan
                    data_out[var].append(val)

        # convert all lists to numpy arrays
        data_out['dtime'] = np.asarray(data_out['dtime'])

        for var in vars_to_read:
            if var in vars_available:
                array = np.asarray(data_out[var])
            else:
                array = np.zeros(len(data_out['dtime'])) * np.nan
            data_out[var] = array

        # compute additional variables (if applicable)
        data_out = self.compute_additional_vars(data_out, vars_to_compute)

        # convert data vectors to pandas.Series (if applicable)
        if vars_as_series:
            for var in (vars_to_read + vars_to_compute):
                if var in vars_to_retrieve:
                    data_out[var] = pd.Series(data_out[var],
                                              index=data_out['dtime'])
                else:
                    del data_out[var]

        return data_out

if __name__ == "__main__":
    read = ReadAeronetSdaV2()

    read.verbosity_level = 'debug'

    first_ten = read.read(last_file=10)

    data = read.read_first_file()
    print(data)



#: Default variable ranges for web display
var_ranges_defaults = {
    "default": {
        "scale": [0, 1.25, 2.5, 3.75, 5, 6.25, 7.5, 8.75, 10],
        "colmap": "coolwarm"
    },
    "ang4487aer": {
        "scale": [0, 0.25, 0.5, 0.75, 1, 1.25, 1.5, 1.75, 2],
        "colmap": "coolwarm"
    },
    "od550aer": {
        "scale": [0, 0.05, 0.1, 0.15, 0.20, 0.25, 0.30, 0.35, 0.40],
        "colmap": "coolwarm"
    },
    "od550lt1aer": {
        "scale": [0, 0.05, 0.1, 0.15, 0.20, 0.25, 0.30, 0.35, 0.40],
        "colmap": "coolwarm"
    },
    "od550gt1aer": {
        "scale": [0, 0.05, 0.1, 0.15, 0.20, 0.25, 0.30, 0.35, 0.40],
        "colmap": "coolwarm"
    },
    "abs550aer": {
        "scale": [0, 0.0125, 0.025, 0.0375, 0.05, 0.0625, 0.075, 0.0875, 0.1],
        "colmap": "coolwarm"
    },
    "absc550aer": {
        "scale": [0, 12.5, 25, 37.5, 50, 62.5, 75, 87.5, 100],
        "colmap": "coolwarm"
    },
    "scatc550dryaer": {
        "scale": [0, 12.5, 25, 37.5, 50, 62.5, 75, 87.5, 100],
        "colmap": "coolwarm"
    },
    "extinction": {
        "scale": [0, 0.125, 0.25, 0.375, 0.5, 0.625, 0.75, 0.875, 1],
        "colmap": "coolwarm"
    },
    "backscatter": {
        "scale": [0, 0.125, 0.25, 0.375, 0.5, 0.625, 0.75, 0.875, 1],
        "colmap": "coolwarm"
    },
    "concso4": {
        "scale": [0, 1.25, 2.5, 3.75, 5, 6.25, 7.5, 8.75, 10],
        "colmap": "coolwarm"
    },
    "concso2": {
        "scale": [0, 1.25, 2.5, 3.75, 5, 6.25, 7.5, 8.75, 10],
        "colmap": "coolwarm"
    },
    "concpm10": {
        "scale": [0, 10, 20, 30, 40, 50, 60, 70, 80],
        "colmap": "coolwarm"
    },
    "concpm25": {
        "scale": [0, 10, 20, 30, 40, 50, 60, 70, 80],
        "colmap": "coolwarm"
    },
    "conco3": {
        "scale": [0, 15, 30, 45, 60, 75, 90, 105, 120],
        "colmap": "coolwarm"
    },
    "vmro3": {
        "scale": [0, 7.5, 15, 22.5, 30, 37.5, 45, 52.5, 60],
        "colmap": "coolwarm"
    },
    "concno2": {
        "scale": [0, 10, 20, 30, 40, 50, 60, 70, 80],
        "colmap": "coolwarm"
    },
    "vmrno2": {
        "scale": [0, 5, 10, 15, 20, 25, 30, 35, 40],
        "colmap": "coolwarm"
    },
    "concNhno3": {
        "scale": [0, 1.25, 2.5, 3.75, 5, 6.25, 7.5, 8.75, 10],
        "colmap": "coolwarm"
    },
    "concNno3pm10": {
        "scale": [0, 1.25, 2.5, 3.75, 5, 6.25, 7.5, 8.75, 10],
        "colmap": "coolwarm"
    },
    "concNno3pm25": {
        "scale": [0, 1.25, 2.5, 3.75, 5, 6.25, 7.5, 8.75, 10],
        "colmap": "coolwarm"
    },
    "concNnh3": {
        "scale": [0, 1.25, 2.5, 3.75, 5, 6.25, 7.5, 8.75, 10],
        "colmap": "coolwarm"
    },
    "concNnh4": {
        "scale": [0, 1.25, 2.5, 3.75, 5, 6.25, 7.5, 8.75, 10],
        "colmap": "coolwarm"
    },
    "concNtno3": {
        "scale": [0, 1.25, 2.5, 3.75, 5, 6.25, 7.5, 8.75, 10],
        "colmap": "coolwarm"
    },
    "concNtnh": {
        "scale": [0, 1.25, 2.5, 3.75, 5, 6.25, 7.5, 8.75, 10],
        "colmap": "coolwarm"
    },
    "concsspm25": {
        "scale": [0, 1.25, 2.5, 3.75, 5, 6.25, 7.5, 8.75, 10],
        "colmap": "coolwarm"
    },
    "concsspm10": {
        "scale": [0, 1.25, 2.5, 3.75, 5, 6.25, 7.5, 8.75, 10],
        "colmap": "coolwarm"
    },
    "concCecpm25": {
        "scale": [0, 1.25, 2.5, 3.75, 5, 6.25, 7.5, 8.75, 10],
        "colmap": "coolwarm"
    },
    "concCocpm25": {
        "scale": [0, 1.25, 2.5, 3.75, 5, 6.25, 7.5, 8.75, 10],
        "colmap": "coolwarm"
    },
    "wetoxs": {
        "scale": [0, 1.25, 2.5, 3.75, 5, 6.25, 7.5, 8.75, 10],
        "colmap": "coolwarm"
    },
    "wetoxn": {
        "scale": [0, 1.25, 2.5, 3.75, 5, 6.25, 7.5, 8.75, 10],
        "colmap": "coolwarm"
    },
    "wetrdn": {
        "scale": [0, 1.25, 2.5, 3.75, 5, 6.25, 7.5, 8.75, 10],
        "colmap": "coolwarm"
    },
    "dryoxs": {
        "scale": [0, 1.25, 2.5, 3.75, 5, 6.25, 7.5, 8.75, 10],
        "colmap": "coolwarm"
    },
    "dryoxn": {
        "scale": [0, 1.25, 2.5, 3.75, 5, 6.25, 7.5, 8.75, 10],
        "colmap": "coolwarm"
    },
    "dryrdn": {
        "scale": [0, 1.25, 2.5, 3.75, 5, 6.25, 7.5, 8.75, 10],
        "colmap": "coolwarm"
    },
}

#: Default information for statistical parameters
statistics_defaults = {
  "nmb": {
    "name": "NMB",
    "longname": "Normalized Mean Bias",
    "scale": [-100, -75, -50, -25, 0, 25, 50, 75, 100],
    "colmap": "bwr",
    "unit": "%",
    "decimals": 1
  },
  "mnmb": {
    "name": "MNMB",
    "longname": "Modified Normalized Mean Bias",
    "scale": [-100, -75, -50, -25, 0, 25, 50, 75, 100],
    "colmap": "bwr",
    "unit": "%",
    "decimals": 1
  },
  "R": {
    "name": "R",
    "longname": "Correlation Coefficient",
    "scale": [0, 0.125, 0.25, 0.375, 0.5, 0.625, 0.75, 0.875, 1],
    "colmap": "RdYlGn",
    "unit": "1",
    "decimals": 2
  },
  "R_spearman": {
    "name": "R Spearman",
    "longname": "R Spearman Correlation",
    "scale": [0, 0.125, 0.25, 0.375, 0.5, 0.625, 0.75, 0.875, 1],
    "colmap": "RdYlGn",
    "unit": "1",
    "decimals": 2
  },
  "fge": {
    "name": "FGE",
    "longname": "Fractional Gross Error",
    "scale": [0, 0.25, 0.5, 0.75, 1, 1.25, 1.5, 1.75, 2],
    "colmap": "reverseColmap(RdYlGn)",
    "unit": "1",
    "decimals": 2
    },
  "nrms": {
    "name": "NRMSE",
    "longname": "Normalized Root Mean Square Error",
    "scale": [0, 25, 50, 75, 100, 125, 150, 175, 200],
    "colmap": "Reds",
    "unit": "%",
    "decimals": 1
  },
  "rms": {
    "name": "RMSE",
    "longname": "Root Mean Square Error",
    "scale": None,
    "colmap": "coolwarm",
    "unit": "1",
    "decimals": 2
  },
  "data_mean": {
    "name": "Mean-Mod",
    "longname": "Model Mean",
    "scale": None,
    "colmap": "coolwarm",
    "unit": "1",
    "decimals": 2
  },
  "refdata_mean": {
    "name": "Mean-Obs",
    "longname": "Observation Mean",
    "scale": None,
    "colmap": "coolwarm",
    "unit": "1",
    "decimals": 2
  },
  

}

# Default information for additional statistical parameters
extended_statistics = {
     "R_spatial_mean": {
        "name": "R-Space",
        "longname": "Spatial R computed from yearly averages",
        "overall_only": True,
        "scale": [0, 0.125, 0.25, 0.375, 0.5, 0.625, 0.75, 0.875, 1],
        "colmap": "RdYlGn",
        "unit": "1",
        "decimals": 2
  }, 
  "R_temporal_median": {
        "name": "R-Temporal",
        "longname": "R temporal median",
        "overall_only": True,
        "scale": [0, 0.125, 0.25, 0.375, 0.5, 0.625, 0.75, 0.875, 1],
        "colmap": "RdYlGn",
        "unit": "1",
        "decimals": 2
  }, 
}

#: Default information about trend display
statistics_trend = {
    "obs/mod_trend": {
        "name": "Obs/Mod-Trends",
        "longname": "Trends",
        "scale": [
            -10.0,
            -7.5,
            -5.0,
            -2.5,
            0,
            2.5,
            5.0,
            7.5,
            10.0
        ],
        "colmap": "bwr",
        "unit": "%/yr",
        "decimals": 1
    },
    "obs_trend": {
        "name": "Obs-Trends",
        "longname": "Observed Trends",
        "scale": [
            -10.0,
            -7.5,
            -5.0,
            -2.5,
            0,
            2.5,
            5.0,
            7.5,
            10.0
        ],
        "colmap": "bwr",
        "unit": "%/yr",
        "decimals": 1
    },
    "mod_trend": {
        "name": "Mod-Trends",
        "longname": "Modelled Trends",
        "scale": [
            -10,
            -7.5,
            -5.0,
            -2.5,
            0,
            2.5,
            5.0,
            7.5,
            10.0
        ],
        "colmap": "bwr",
        "unit": "%/yr",
        "decimals": 1
    },
}

#: Mapping of pyaerocom variable names to web naming conventions
var_web_info = dict(

    od550aer      = ['AOD', '2D', 'Optical properties'],
    od550csaer    = ['AOD (clear sky)', '2D', 'Optical properties'],
    od550lt1aer   = ['AODf', '2D', 'Optical properties'],
    od550gt1aer   = ['AODc', '2D', 'Optical properties'],
    abs550aer     = ['AAOD', '2D', 'Optical properties'],
    ang4487aer    = ['AE', '2D', 'Optical properties'],
    angabs4487aer = ['AAE', '2D', 'Optical properties'],
    ang4487csaer  = ['AE (clear sky)', '2D', 'Optical properties'],
    sc550dryaer   = ['Scat. coef. (dry)', '3D', 'Optical properties'],
    sc550aer      = ['Scat. coef.', '3D', 'Optical properties'],
    ac550aer      = ['Abs. coef.', '3D', 'Optical properties'],
    ac550dryaer   = ['Abs. coef. (dry)', '3D', 'Optical properties'],
    ec532aer      = ['Ext. coeff.', '3D', 'Optical properties'],
    bsc532aer     = ['Backscat. coeff.', '3D', 'Optical properties'],
    concpm10      = ['PM10', '3D', 'Particle concentrations'],
    concpm25      = ['PM2.5', '3D', 'Particle concentrations'],
    concso4       = ['SO4', '3D', 'Particle concentrations'],

    concso4pr     = ['SO4 (precip.)', '3D', 'Particle concentrations'],
    concbc        = ['BC', '3D', 'Particle concentrations'],
    concoa        = ['OA', '3D', 'Particle concentrations'],
    concss        = ['SS', '3D', 'Particle concentrations'],
    conco3        = ['O3', '3D', 'Gas concentrations'],
    concso2       = ['SO2', '3D', 'Gas concentrations'],
    vmrox         = ['OX', '3D', 'Gas volume mixing ratio'],
    vmrso2        = ['SO2', '3D', 'Gas volume mixing ratio'],
    conctno3      = ['tNO3', '3D', 'Concentration'],
    conctnh       = ['tNH', '3D', 'Concentration'],
    concnh3       = ['NH3', '3D', 'Concentration'],
    conchno3      = ['HNO3', '3D', 'Concentration'],
    concno310     = ['NO3_PM10', '3D', 'Particle concentration'],
    concno325     = ['NO3_PM25', '3D', 'Particle concentration'],
    concss10      = ['SS_PM10', '3D', 'Particle concentration'],
    concss25      = ['SS_PM25', '3D', 'Particle concentration'],
    concec        = ['EC', '3D', 'Particle concentration'],
    conccoc       = ['OC', '3D', 'Particle concentration'],
    wetoxs        = ['WetOXS', '3D', 'Deposition'],
    drysox        = ['DryOXS', '3D', 'Deposition'],
    concnh4       = ['NH4', '3D', 'Gas concentrations'],
    concno3       = ['NO3', '3D', 'Gas concentrations'],
    vmro3         = ['O3', '3D', 'Volume mixing ratios'],
    concNtno3      = ['tNO3', '3D', 'Concentration'],
    concNtnh       = ['tNH', '3D', 'Concentration'],
    concNnh3       = ['NH3', '3D', 'Concentration'],
    concNhno3      = ['HNO3', '3D', 'Concentration'],
    concNno3pm10     = ['NO3_PM10', '3D', 'Particle concentration'],
    concNno3pm25     = ['NO3_PM25', '3D', 'Particle concentration'],
    concsspm10      = ['SS_PM10', '3D', 'Particle concentration'],
    concsspm25      = ['SS_PM25', '3D', 'Particle concentration'],
    concCecpm25        = ['EC', '3D', 'Particle concentration'],
    concCocpm25       = ['OC', '3D', 'Particle concentration'],
    wetoxn        = ['WetOXN', '3D', 'Deposition'],
    wetrdn        = ['WetRDN', '3D', 'Deposition'],
    dryoxs        = ['FakeDryOXS', '3D', 'Deposition'],
    dryoxn        = ['FakeDryOXN', '3D', 'Deposition'],
    dryrdn        = ['FakeDryRDN', '3D', 'Deposition'],
    concno2       = ['NO2', '3D', 'Gas concentrations'],
    concNnh4       = ['NH4', '3D', 'Gas concentrations'],
    vmrco         = ['CO', '3D', 'Volume mixing ratios'],
    vmrno2        = ['NO2', '3D', 'Volume mixing ratios'],
    concco        = ['CO', '3D', 'Particle concentration'],
)



"""
Definitions and helpers related to time conversion
"""

from iris import coord_categorisation
from datetime import datetime

TS_TYPES = ['minutely', 'hourly', 'daily', 'weekly', 'monthly', 'yearly',
            'native']

# The following import was removed and the information about available unit
# strings was copied from the netCDF4 module directly here
microsec_units = ['microseconds', 'microsecond', 'microsec', 'microsecs']
millisec_units = ['milliseconds', 'millisecond', 'millisec', 'millisecs']
sec_units = ['second', 'seconds', 'sec', 'secs', 's']
min_units = ['minute', 'minutes', 'min', 'mins']
hr_units = ['hour', 'hours', 'hr', 'hrs', 'h']
day_units = ['day', 'days', 'd']

#
# Start of the gregorian calendar
# adapted from here: https://github.com/Unidata/cftime/blob/master/cftime/_cftime.pyx
GREGORIAN_BASE = datetime(1582, 10, 15)

IRIS_AGGREGATORS = {'hourly'    :   coord_categorisation.add_hour,
                    'daily'     :   coord_categorisation.add_day_of_year,
                    'monthly'   :   coord_categorisation.add_month_number,
                    'yearly'    :   coord_categorisation.add_year}

# some helper dictionaries for conversion of temporal resolution
# https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#offset-aliases
TS_TYPE_TO_PANDAS_FREQ = {'minutely':   'T',
                          'hourly'  :   'H',
                          'daily'   :   'D',
                          'weekly'  :   'W-MON',
                          'monthly' :   'MS', #Month start !
                          'season'  :   'Q',
                          'yearly'  :   'AS'}

PANDAS_RESAMPLE_OFFSETS = {'AS' : '181D',
                           'MS' : '14D'}

PANDAS_FREQ_TO_TS_TYPE = {v: k for k, v in TS_TYPE_TO_PANDAS_FREQ.items()}

# frequency strings
# https://numpy.org/devdocs/reference/arrays.datetime.html#datetime-units
TS_TYPE_TO_NUMPY_FREQ =  {'minutely':   'm',
                          'hourly'  :   'h',
                          'daily'   :   'D',
                          'weekly'  :   'W',
                          'monthly' :   'M', #Month start !
                          'yearly'  :   'Y'}

# conversion of ts_types to strings that cf_units understands
TS_TYPE_TO_SI = {
    'minutely':     'min',
    'hourly'  :     'h',
    'daily'   :     'd',
    'weekly'  :     'week',
    'monthly' :     'month',
    'yearly'  :     'yr'
}

SI_TO_TS_TYPE = {v: k for k, v in TS_TYPE_TO_SI.items()}

# conversion of datetime-like objects for given temporal resolutions (can, e.g.
# be used in plotting methods)
TS_TYPE_DATETIME_CONV = {None       : '%d.%m.%Y', #Default
                         'hourly'   : '%d.%m.%Y',
                         '3hourly'  : '%d.%m.%Y',
                         'daily'    : '%d.%m.%Y',
                         'weekly'   : '%d.%m.%Y',
                         'monthly'  : '%b %Y',
                         'yearly'   : '%Y'}

TS_TYPE_SECS = {'minutely': 60,
                'hourly'  : 3600,
                '3hourly' : 10800,
                'daily'   : 86400,
                'weekly'  : 604800,
                'monthly' : 2592000, #counting 3 days per month (APPROX)
                'yearly'  : 31536000} #counting 365 days (APPROX)


import iris
import numpy as np
from traceback import format_exc
from pyaerocom import print_log
from pyaerocom._lowlevel_helpers import merge_dicts
from pyaerocom.helpers import copy_coords_cube
from pyaerocom.molmasses import get_molmass
from pyaerocom.units_helpers import get_unit_conversion_fac
from pyaerocom.molmasses import get_mmr_to_vmr_fac


CUBE_MATHS = {'add'         :   iris.analysis.maths.add,
              'subtract'    :   iris.analysis.maths.subtract,
              'divide'      :   iris.analysis.maths.divide,
              'multiply'    :   iris.analysis.maths.multiply}

def _apply_operator_cubes(cube1, cube2, operator_name,
                          allow_coord_merge=True):
    if not operator_name in CUBE_MATHS:
        raise NotImplementedError('No such arithmetic cube operator '
                                  'implemented: {}'.format(operator_name))
    fun = CUBE_MATHS[operator_name]
    try:
        return fun(cube1, cube2)
    except ValueError as e:
        print_log.warning('Could not {} cubes straight out of the box. Trying '
                          'to correct for dimension definition errors'
                          .format(operator_name))
        if 'differing coordinates (time)' in repr(e):
            iris.util.unify_time_units([cube1, cube2])
            attrs = {}
            attrs.update(cube1.coord('time').attributes)
            attrs.update(cube2.coord('time').attributes)
            cube2.coord('time').attributes = attrs
            cube1.coord('time').attributes = attrs

        elif allow_coord_merge:
            copy_coords_cube(to_cube=cube2,
                             from_cube=cube1,
                             inplace=True)
    return fun(cube1, cube2)

def _check_input_iscube(*data_objs):
    from pyaerocom.griddeddata import GriddedData
    checked = []
    for obj in data_objs:
        if isinstance(obj, GriddedData):
            checked.append(obj.cube)
        elif isinstance(obj, iris.cube.Cube):
            checked.append(obj)
        else:
            raise ValueError('Invalid input: require GriddedData or Cube, got '
                             '{}'.format(type(obj)))
    return checked

def _check_same_units(cube1, cube2):
    if cube1.units == cube2.units:
        return (cube1, cube2)
    from pyaerocom import const
    var1 = cube1.var_name
    u1 = cube1.units

    var2 = cube2.var_name
    u2 = cube2.units

    if var1 in const.VARS and u1 == const.VARS[var1]['units'] :
        cube2.convert_units(u1)
        return (cube1, cube2)
    elif var2 in const.VARS and u2 == const.VARS[var2]['units'] :
        cube1.convert_units(u2)
        return (cube1, cube2)

    try:
        cube2.convert_units(u1)
        return (cube1, cube2)
    except Exception:
        from pyaerocom.exceptions import UnitConversionError
        raise UnitConversionError('Failed to harmonise units')

def merge_meta_cubes(cube1, cube2):
    try:
        return merge_dicts(cube1.attributes,
                           cube2.attributes)
    except Exception:
        print_log.warning('WARNING: Failed to merge Cube metadata. Reason:\n{}'
                          .format(format_exc()))
        ts_type = None
        try:
            if cube1.attributes['ts_type']==cube2.attributes['ts_type']:
                ts_type = cube1.attributes['ts_type']
        except:
            pass

        return {'NOTE'      : 'METADATA_MERGE_FAILED',
                'meta1'     : cube1.attributes,
                'meta2'     : cube2.attributes,
                'ts_type'   : ts_type}

def apply_rh_thresh_cubes(cube, rh_cube, rh_max=None):
    """Method that applies a low RH filter to input cube"""
    cube, rh_cube =  _check_input_iscube(cube, rh_cube)
    if rh_max is None:
        from pyaerocom import const
        rh_max = const.VARS[cube.var_name]['dry_rh_max']
    if not cube.shape == rh_cube.shape:
        raise ValueError
    mask = rh_cube.data > rh_max
    cube_out = cube.copy()
    cube_out.data[mask] = np.nan
    cube_out.attributes.update(cube.attributes)
    cube_out.attributes['rh_max'] = rh_max
    return cube_out

def add_cubes(cube1, cube2):
    """Method to add cubes from 2 gridded data objects
    """
    cube1, cube2 = _check_input_iscube(cube1, cube2)
    cube1, cube2 = _check_same_units(cube1, cube2)

    cube_out = _apply_operator_cubes(cube1, cube2, 'add',
                                     allow_coord_merge=True)

    cube_out.attributes.update(merge_meta_cubes(cube1, cube2))
    return cube_out

def subtract_cubes(cube1, cube2):
    """Method to subtract 1 cube from another"""
    cube1, cube2 = _check_input_iscube(cube1, cube2)
    cube1, cube2 = _check_same_units(cube1, cube2)

    cube_out = _apply_operator_cubes(cube1, cube2, 'subtract',
                                     allow_coord_merge=True)

    cube_out.attributes.update(merge_meta_cubes(cube1, cube2))
    return cube_out

def multiply_cubes(cube1, cube2):
    """Method to multiply 2 cubes"""
    cube1, cube2 = _check_input_iscube(cube1, cube2)
    cube_out = _apply_operator_cubes(cube1, cube2, 'multiply',
                                     allow_coord_merge=True)

    cube_out.attributes.update(merge_meta_cubes(cube1, cube2))
    return cube_out

def divide_cubes(cube1, cube2):
    """Method to divide 2 cubes with each other"""
    cube1, cube2 = _check_input_iscube(cube1, cube2)
    cube_out = _apply_operator_cubes(cube1, cube2, 'divide',
                                     allow_coord_merge=True)

    cube_out.attributes.update(merge_meta_cubes(cube1, cube2))
    return cube_out

def lifetime_from_load_and_dep(load, wetdep, drydep):
    """Compute lifetime from load and wet and dry deposition"""
    raise NotImplementedError('Lifetime cannot be computed based on '
                              'grid data')
    load, wetdep, drydep = _check_input_iscube(load, wetdep, drydep)
    deptot = _apply_operator_cubes(wetdep, drydep, 'add',
                                   allow_coord_merge=True)

    deptot.attributes.update(merge_meta_cubes(wetdep, drydep))

    cube_out = _apply_operator_cubes(load, deptot, 'divide',
                                     allow_coord_merge=True)
    cube_out.attributes.update(merge_meta_cubes(load, deptot))
    return cube_out

def compute_angstrom_coeff_cubes(cube1, cube2, lambda1=None, lambda2=None):
    """Compute Angstrom coefficient cube based on 2 optical densitiy cubes

    Parameters
    ----------
    cube1 : iris.cube.Cube
        AOD at wavelength 1
    cube2 : iris.cube.Cube
        AOD at wavelength 2
    lambda1 : float
        wavelength 1
    lambda 2 : float
        wavelength 2

    Returns
    -------
    Cube
        Cube containing Angstrom exponent(s)
    """
    from pyaerocom.varnameinfo import VarNameInfo
    from cf_units import Unit
    cube1, cube2 = _check_input_iscube(cube1, cube2)
    if lambda1 is None:
        lambda1 = VarNameInfo(cube1.var_name).wavelength_nm
    if lambda2 is None:
        lambda2 = VarNameInfo(cube2.var_name).wavelength_nm

    ratio = _apply_operator_cubes(cube1, cube2, 'divide')
    logr = iris.analysis.maths.log(ratio)

    wvl_r = np.log(lambda1 / lambda2)
    cube_out = -1*iris.analysis.maths.divide(logr, wvl_r)
    cube_out.units = Unit('1')
    cube_out.attributes.update(merge_meta_cubes(cube1, cube2))
    return cube_out


def rho_from_ts_ps(ts, ps):
    R = 287.058 #R for dry air

    rho = R*divide_cubes(ts,ps)

    rho.attributes.update(merge_meta_cubes(ts, ps))

    return rho

def mmr_from_vmr(cube):
    """
    Convvert gas volume/mole mixing ratios into mass mixing ratios.

    Parameters
    ----------
    cube : iris.cube.Cube
        A cube containing gas vmr data to be converted into mmr.

    Returns
    -------
    cube_out : iris.cube.Cube
        Cube containing mmr data.

    """
    cube = _check_input_iscube(cube)[0]

    var_name = cube.var_name
    M_dry_air = get_molmass('air_dry')
    M_variable = get_molmass(var_name)

    cube_out = (M_variable/M_dry_air)*cube
    return cube_out

def conc_from_vmr(cube, ts, ps):
    cube,ts,ps = _check_input_iscube(cube,ts,ps)

    mmr_cube = mmr_from_vmr(cube)
    rho = rho_from_ts_ps(ts, ps)

    cube_out = multiply_cubes(mmr_cube,rho)

    return cube_out

def conc_from_vmr_STP(cube,):
    cube = _check_input_iscube(cube)[0]

    R = 287.058 #R for dry air

    standard_T = 293
    standard_P = 101300


    mmr_cube = mmr_from_vmr(cube)
    rho = R*standard_T/standard_P

    cube_out = rho*mmr_cube

    return cube_out


def mmr_to_vmr_cube(data):
    """
    Convert cube containing MMR data to VMR

    Parameters
    ----------
    data : iris.Cube or GriddedData
        input data object containing MMR data for a certain variable. Needs
        to have `var_name` attr. assigned and valid MMR AeroCom variable name
        (e.g. mmro3, mmrno2)

    Raises
    ------
    AttributeError
        if attr. `var_name` of input data does not start with mmr

    Returns
    -------
    iris.Cube
        cube containing mixing ratios expressed as VMR in units of nmole mole-1

    """
    from pyaerocom import GriddedData
    var = data.var_name
    if not var.startswith('mmr'):
        raise AttributeError(
            f'MMR to VMR can only be done for MMR variables (which must start '
            f'with mmr, however, the var_name of the input data is {var}')

    # make sure mmr is in correct units (the function will crash if not)
    get_unit_conversion_fac(data.units, 'kg kg-1')
    if isinstance(data, GriddedData):
        data = data.cube

    vmrvar = var.replace('mmr', 'vmr')

    mulfac = get_mmr_to_vmr_fac(var) * 1e9 # 1e9 is to go to nmole mole-1

    cube = data * mulfac
    cube.attributes.update(data.attributes)
    cube.units = 'nmole mole-1'
    cube.var_name = vmrvar
    return cube




"""
This module contains scatter plot routines for Aerocom data.
"""
from matplotlib.ticker import ScalarFormatter
import os
import matplotlib.pyplot as plt
import numpy as np
from pyaerocom import const
from pyaerocom.helpers import start_stop_str
from pyaerocom.mathutils import calc_statistics, exponent

def plot_scatter(x_vals, y_vals, **kwargs):
    """Scatter plot

    Currently a wrapper for high-level method plot_scatter_aerocom (same module,
    see there for details)
    """
    return plot_scatter_aerocom(x_vals, y_vals, **kwargs)

def plot_scatter_aerocom(x_vals, y_vals, var_name=None, var_name_ref=None,
                         x_name=None, y_name=None, start=None, stop=None,
                         ts_type=None, unit=None, stations_ok=None,
                         filter_name=None, lowlim_stats=None,
                         highlim_stats=None, loglog=None, ax=None, figsize=None,
                         fontsize_base=11, fontsize_annot=None,
                         marker=None, color=None, alpha=0.5,
                         **kwargs):
    """Method that performs a scatter plot of data in AEROCOM format

    Parameters
    ----------
    y_vals : ndarray
        1D array (or list) of model data points (y-axis)
    x_vals : ndarray
        1D array (or list) of observation data points (x-axis)
    var_name : :obj:`str`, optional
        name of variable that is plotted
    var_name_ref : :obj:`str`, optional
        name of variable of reference data
    x_name : :obj:`str`, optional
        Name of observation network
    y_name : :obj:`str`, optional
        Name / ID of model
    start : :obj:`str` or :obj`datetime` or similar
        start time of data
    stop : :obj:`str` or :obj`datetime` or similar
        stop time of data
    ts_type : str
        frequency of data
    unit : str, optional
        unit of data
    stations_ok : int, optional
        number of stations from which data were generated
    filter_name : str, optional
        name of filter
    lowlim_stats : float, optional
        lower value considered for statistical parameters
    highlim_stats : float, optional
        upper value considered for statistical parameters
    loglog : bool, optional
        plot log log scale, if None, pyaerocom default is used
    ax : Axes
        axes into which the data are to be plotted
    figsize : tuple
        size of figure (if new figure is created, ie ax is None)
    fontsize_base : int
        basic fontsize, defaults to 11
    fontsize_annot : int, optional
        fontsize used for annotations
    marker : str, optional
        marker used for data, if None, '+' is used
    color : str, optional
        color of markers, default to 'k'
    alpha : float, optional
        transparency of markers (does not apply to all marker types),
        defaults to 0.5.
    **kwargs
        additional keyword args passed to :func:`ax.plot`


    Returns
    -------
    matplotlib.axes.Axes
        plot axes
    """
    if marker is None:
        marker='+'
    if color is None:
        color = 'k'
    if isinstance(y_vals, list):
        y_vals = np.asarray(y_vals)
    if isinstance(x_vals, list):
        x_vals = np.asarray(x_vals)
    try:
        var = const.VARS[var_name]
    except Exception:
        var = const.VARS['DEFAULT']

    try:
        var_ref = const.VARS[var_name_ref]
    except Exception:
        var_ref = const.VARS['DEFAULT']

    if loglog is None:
        loglog = var_ref.scat_loglog

    xlim = var['scat_xlim']
    ylim = var_ref['scat_ylim']

    if xlim is None or ylim is None:
        low  =  np.min([np.nanmin(x_vals), np.nanmin(y_vals)])
        high =  np.max([np.nanmax(x_vals), np.nanmax(y_vals)])

        xlim = [low, high]
        ylim = [low, high]

    if ax is None:
        if figsize is None:
            figsize = (10,8)
        fig, ax = plt.subplots(figsize=figsize)
    if var_name is None:
        var_name = 'n/d'

    statistics = calc_statistics(y_vals, x_vals,
                                 lowlim=lowlim_stats,
                                 highlim=highlim_stats)

    if loglog:
        ax.loglog(x_vals, y_vals, ls='none', color=color, marker=marker,
                  alpha=alpha, **kwargs)
    else:
        ax.plot(x_vals, y_vals, ls='none', color=color, marker=marker,
                  alpha=alpha, **kwargs)

    try:
        title = start_stop_str(start, stop, ts_type)
        if ts_type is not None:
            title += ' ({})'.format(ts_type)
    except Exception:
        title = ''

    if not loglog:
        xlim[0] = 0
        ylim[0] = 0
    elif any(x[0] < 0 for x in [xlim, ylim]):
        low = np.nanmin(y_vals)
        if low != 0:
            low = 10**(float(exponent(abs(low)) - 1))
        xlim[0] = low
        ylim[0] = low
    ax.set_xlim(xlim)
    ax.set_ylim(ylim)
    xlbl = '{}'.format(x_name)
    if var_name_ref is not None:
        xlbl += ' ({})'.format(var_name_ref)
    ax.set_xlabel(xlbl, fontsize=fontsize_base+4)
    ax.set_ylabel('{}'.format(y_name), fontsize=fontsize_base+4)

    ax.set_title(title, fontsize=fontsize_base+4)

    ax.tick_params(labelsize=fontsize_base)

    ax.plot(xlim, ylim, '--', color='#cccccc')

    xypos =   {'var_info'       :   (0.01, .95),
               'refdata_mean'   :   (0.01, 0.90),
               'data_mean'      :   (0.01, 0.86),
               'nmb'            :   (0.01, 0.82),
               'mnmb'           :   (0.35, 0.82),
               'R'              :   (0.01, 0.78),
               'rms'            :   (0.35, 0.78),
               'R_kendall'      :   (0.01, 0.74),
               'fge'            :   (0.35, 0.74),
               'ts_type'        :   (0.8, 0.1),
               'filter_name'    :   (0.8, 0.06)}

    var_str = var_name
    _ndig = abs(exponent(statistics['refdata_mean']) - 2)
    if unit is None:
        unit = 'N/D'
    if not str(unit) in ['1', 'no_unit']:
        var_str += ' [{}]'.format(unit)

    if fontsize_annot is None:
        fontsize_annot = fontsize_base
    ax.annotate("{} #: {} # st: {}".format(var_str,
                        statistics['num_valid'], stations_ok),
                        xy=xypos['var_info'], xycoords='axes fraction',
                        fontsize=fontsize_annot+4, color='red')

    ax.annotate('Mean (x-data): {:.{}f}; Rng: [{:.{}f}, {:.{}f}]'
                .format(statistics['refdata_mean'], _ndig,
                        np.nanmin(x_vals),_ndig,
                        np.nanmax(x_vals), _ndig),
                xy=xypos['refdata_mean'], xycoords='axes fraction',
                fontsize=fontsize_annot,
                color='red')

    ax.annotate('Mean (y-data): {:.{}f}; Rng: [{:.{}f}, {:.{}f}]'
                .format(statistics['data_mean'], _ndig,
                        np.nanmin(y_vals),_ndig,
                        np.nanmax(y_vals), _ndig),
                xy=xypos['data_mean'], xycoords='axes fraction',
                fontsize=fontsize_annot,
                color='red')

    ax.annotate('NMB: {:.1f}%'.format(statistics['nmb']*100),
                        xy=xypos['nmb'], xycoords='axes fraction',
                        fontsize=fontsize_annot, color='red')

    ax.annotate('MNMB: {:.1f}%'.format(statistics['mnmb']*100),
                        xy=xypos['mnmb'], xycoords='axes fraction',
                        fontsize=fontsize_annot, color='red')

    ax.annotate('R (Pearson): {:.3f}'.format(statistics['R']),
                        xy=xypos['R'], xycoords='axes fraction',
                        fontsize=fontsize_annot, color='red')

    ax.annotate('RMS: {:.3f}'.format(statistics['rms']),
                        xy=xypos['rms'], xycoords='axes fraction',
                        fontsize=fontsize_annot, color='red')

    ax.annotate('R (Kendall): {:.3f}'.format(statistics['R_kendall']),
                        xy=xypos['R_kendall'], xycoords='axes fraction',
                        fontsize=fontsize_annot, color='red')

    ax.annotate('FGE: {:.1f}'.format(statistics['fge']),
                        xy=xypos['fge'], xycoords='axes fraction',
                        fontsize=fontsize_annot, color='red')
    # right lower part
    ax.annotate('{}'.format(ts_type),
                        xy=xypos['ts_type'], xycoords='axes fraction',
                        ha='center',
                        fontsize=fontsize_annot, color='black')
    ax.annotate('{}'.format(filter_name),
                        xy=xypos['filter_name'], xycoords='axes fraction',
                        ha='center',
                        fontsize=fontsize_annot, color='black')

    ax.set_aspect('equal')
    return ax



from . import config
from . import mapping
from . import plotscatter
from . import plotcoordinates
from . import heatmaps

from .mapping import plot_griddeddata_on_map, init_map
from .plotscatter import plot_scatter
from .heatmaps import df_to_heatmap
from .plotcoordinates import plot_coordinates


from scipy.constants import Avogadro
import numpy as np

def mass_to_nr_molecules(mass, mm):
    """ Calculating the number of molecules form mass and molarmass.

    Mass, Molar mass need to be in the same unit, either both g and g/mol
    or kg and kg/mol.

    Parameters
    ---------------------------
    mass : float
        mass of all compounds.

    mm : float
        molar mass of compounds.

    Returns
    -----------------
    nr_molecules : float
        number of molecules
    """
    nr_molecules = mass/mm*Avogadro
    return nr_molecules

def nr_molecules_to_mass(nr_molecules, mm):
    """ Calculates the mass from the number of molecules and molar mass.

    Parameters
    ---------------
    nr_molecules : int
        Number of molecules

    mm : float
        Molar mass [g/mol]

    Returns
    ---------------------
    mass : float
        mass in grams
    """
    mass = mm*nr_molecules/Avogadro
    return mass

def unitconv_sfc_conc_bck(data, x = 2):
    """ Converting: ugSOx/m3 to  ugS/ m3.

    Parameters
    ------------------
    data: ndarray
        Contains the data in units of ugSoX/m3.

    x: int
        The number of oxygen atoms, O in you desired SOx compound.

    Returns
    ------------
    data : ndarray
        in units of ugS/ m3.

    Notes
    -----------
    micro grams to kilos is 10**6
    """

    mmO = 15.9999 # molar mass oxygen
    mmS = 32.065  # molar mass sulphur
    mm_compound =  (mmS + x*mmO)*10**3 # *10**3 gives molar mass in micrograms

    nr_molecules = mass_to_nr_molecules(data, mm_compound)
    weight_s = nr_molecules_to_mass(nr_molecules, mmS*10**3) # weigth in ug
    return weight_s

def unitconv_sfc_conc(data, nr_of_O = 2):
    """ Unitconverting:  ugS/m3 to ugSOx/m3

    Parameters
    ------------------
    data : array_like
        Contains the data in units of ugS/m3.

    nr_of_O: int
        The number of O's in you desired SOx compound.

    Returns
    ------------
    data : ndarray
        data in units of ug SOx/m3

    """

    mm_s = 32.065*10**6 # in units of ug/mol
    mm_o = nr_of_O*15.9999*10**6 ## in units of ug/mol
    nr_molecules = mass_to_nr_molecules(data, mm_s) # 32.065*10**6) [ug/mol]
    added_weight_oksygen = nr_molecules_to_mass(nr_molecules, mm_o) # ug
    # added weights in micrograms
    mass = data + added_weight_oksygen # in micrograms
    return mass

def unitconv_wet_depo_bck(data, time, ts_type = "monthly"):
    """ The unitconversion kg SO4 m-2 s-1 to kgS/ha.

    Removing the weight of oxygen.

    Parameters
    ------------------
    data: ndarray
        Sulphur data you wish to convert.

    time : pd.Seires[numpy.datetime64]
         Array of datetime64 timesteps.

    ts_type: str
       The timeseries type. Default monthly.

    Returns
    ------------------
    data : ndarray
       Sulphur data in units of ugSOx m-3 s-1.

    """
    # kg SO4 m-2 s-1 to kg S/ha
    mm_so4 = 0.001*32.065 + 0.001*15.999*4 # kg/mol
    mm_s = 32.065*0.001 # kg/mol
    mm_o = 0.001*15.999*4 # kg/mol

    days_in_month = time.dt.daysinmonth
    monthly_to_sec = days_in_month*24*60*60 # Seconds in each
    nr_molecules = data*Avogadro/mm_so4 # [1]
    mass_S       = nr_molecules*mm_s/Avogadro # mass in kg
    # Mulitply by seconds in one month
    mass_pr_ha = mass_S*monthly_to_sec*10000
    return mass_pr_ha

def unitconv_wet_depo(data, time, ts_type = "monthly"):
    """ Unitconversion kg S/ha to kg SOx m-2 s-1.

    Adding mass of oksygen.

    Parameters
    ------------------
    data: ndarray
        data in unit kg S/ha = kg S/(1000 m2)

    time : pd.Seires[numpy.datetime64]
        Array of datetime64 timesteps.

    ts_type : str
        The timeseries type. Default "monthly".

    Returns
    ------------------
    data : ndarray
        data in units of ugSOx/m3

    """
    mmSO4 = 0.001*32.065 + 0.001*15.999*4 # in kg/mol
    mm_s = 32.065*0.001 # kg/mol
    #print('uses new updated version, in that case fix conversion')
    nr_molecules = data*Avogadro/mm_s # [1]
    mass_SO4 = nr_molecules*mmSO4/Avogadro # mass in kl
    days_in_month = time.dt.daysinmonth
    monthly_to_sec = days_in_month*24*60*60
    #print('includes new changes')
    mass_pr_square_m_pr_sek = mass_SO4/(10000*monthly_to_sec)
    return mass_pr_square_m_pr_sek

def unitconv_wet_depo_from_emep(data, time, ts_type = "monthly"):
    """ Unitconversion mgS m-2 to kg SO4 m-2 s-1.

    Milligram to kilos is 10-6.

    Adding mass of oksygen.

    Parameters
    ------------------
    data: ndarray
        data in unit mg S m-2.

    time : pd.Seires[numpy.datetime64]
        Array of datetime64 timesteps.

    ts_type : str
        The timeseries type. Default "monthly".

    Returns
    ------------------
    data : ndarray
        data in units of ugSOx/m3

    """
    # TODO add if time is not of correct pandas series convert
    # numpy ndarray to pandas series. Much better than having to remeber that
    #the only thing thats a åandas seies.
    # If time

    mm_so4 = 0.001*32.065 + 0.001*15.999*4 # in kg/mol
    mm_s  = 0.001*32.065 # kg/mol
    data_in_kilos = data*10**(-9)
    nr_molecules = data_in_kilos*Avogadro/mm_s  # [1]
    mass_SO4     = nr_molecules*mm_so4/Avogadro # mass in kg
    days_in_month  = time.dt.daysinmonth
    monthly_to_sec = days_in_month*24*60*60
    mass_pr_square_m_pr_sek = mass_SO4*10000/monthly_to_sec
    return mass_pr_square_m_pr_sek



from ast import literal_eval
from configparser import ConfigParser
import numpy as np

from pyaerocom import logger, print_log, var_groups
from pyaerocom.mathutils import make_binlist
from pyaerocom.obs_io import OBS_WAVELENGTH_TOL_NM
from pyaerocom.exceptions import VariableDefinitionError
from pyaerocom._lowlevel_helpers import list_to_shortstr, dict_to_str

#: helper vor checking if variable name contains str 3d or 3D
from pyaerocom.variable_helpers import parse_variables_ini, parse_aliases_ini, \
    _read_alias_ini, get_aliases, _check_alias_family
from pyaerocom.varnameinfo import VarNameInfo

class Variable(object):
    """Interface that specifies default settings for a variable

    See `variables.ini <https://github.com/metno/pyaerocom/blob/master/
    pyaerocom/data/variables.ini>`__ file for an overview of currently available
    default variables.

    Parameters
    ----------
    var_name : str
        string ID of variable (see file variables.ini for valid IDs)
    init : bool
        if True, input variable name is attempted to be read from config file
    cfg : ConfigParser
        open config parser that holds the information in config file available
        (i.e. :func:`ConfigParser.read` has been called with config file as
        input)
    **kwargs
        any valid class attribute (e.g. map_vmin, map_vmax, ...)

    Attributes
    ----------
    var_name : str
        input variable name
    var_name_aerocom : str
        AEROCOM variable name (see e.g. `AEROCOM protocol
        <http://aerocom.met.no/protocol_table.html>`__ for a list of
        available variables)
    is_3d : bool
        flag that indicates if variable is 3D
    is_dry : bool
        flag that is set based on filename that indicates if variable data
        corresponds to dry conditions.
    units : str
        unit of variable (None if no unit)
    default_vert_code : str, optional
        default vertical code to be loaded (i.e. Column, ModelLevel, Surface).
        Only relevant during reading and in case conflicts occur (e.g.
        abs550aer, 2010, Column and Surface files)
    aliases : list
        list of alternative names for this variable
    minimum : float
        lower limit of allowed value range
    upper_limit : float
        upper limit of allowed value range
    obs_wavelength_tol_nm : float
        wavelength tolerance (+/-) for reading of obsdata. Default is 10, i.e.
        if this variable is defined at 550 nm and obsdata contains measured
        values of this quantity within interval of 540 - 560, then these data
        is used
    scat_xlim : float
        x-range for scatter plot
    scat_ylim : float
        y-range for scatter plot
    scat_loglog : bool
        scatter plot on loglog scale
    scat_scale_factor : float
        scale factor for scatter plot
    map_cmap : str
        name of default colormap (matplotlib) of this variable.
    map_vmin : float
        data value corresponding to lower end of colormap in map plots of this
        quantity
    map_vmax : float
        data value corresponding to upper end of colormap in map plots of this
        quantity
    map_c_under : str
        color used for values below :attr:`map_vmin` in map plots of this
        quantity
    map_c_over : str
        color used for values exceeding :attr:`map_vmax` in map plots of this
        quantity
    map_cbar_levels : :obj:`list`, optional
        levels of colorbar
    map_cbar_ticks : :obj:`list`, optional
        colorbar ticks
    """
    literal_eval_list = lambda val: list(literal_eval(val))
    str2list = lambda val: [x.strip() for x in val.split(',')]
    str2bool = lambda val: val.lower() in ('true', '1', 't', 'yes')

    _TYPE_CONV={
        'wavelength_nm': float,
        'minimum': float,
        'maximum': float,
        'dimensions' : str2list,
        'obs_wavelength_tol_nm': float,
        'scat_xlim': literal_eval_list,
        'scat_ylim': literal_eval_list,
        'scat_loglog': str2bool,
        'scat_scale_factor': float,
        'dry_rh_max':float,
        'map_cmap' : str,
        'map_vmin': float,
        'map_vmax': float,
        'map_cbar_levels': literal_eval_list,
        'map_cbar_ticks': literal_eval_list,
        '_is_rate'      : bool
        }

    #maybe used in config
    ALT_NAMES = {'unit' : 'units'}

    plot_info_keys = ['scat_xlim',
                      'scat_ylim',
                      'scat_loglog',
                      'scat_scale_factor',
                      'map_vmin',
                      'map_vmax',
                      'map_cmap',
                      'map_c_under',
                      'map_c_over',
                      'map_cbar_levels',
                      'map_cbar_ticks']
    VMIN_DEFAULT = -np.inf
    VMAX_DEFAULT = np.inf

    @staticmethod
    def _check_input_var_name(var_name):
        if '3d' in var_name:
            var_name = var_name.replace('3d','')
        elif '3D' in var_name:
            var_name = var_name.replace('3D', '')
        elif '_' in var_name:
            raise ValueError(f'invalid variable name {var_name}. Must '
                             f'not contain underscore')
        return var_name

    def __init__(self, var_name=None, init=True, cfg=None, **kwargs):
        if var_name is None:
            var_name = 'od550aer'
        elif not isinstance(var_name, str):
            raise ValueError(f'Invalid input for variable name, need str '
                             f'type, got {type(var_name)}')
        # save orig. input for whatever reason
        self._var_name_input = var_name

        self.var_name = self._check_input_var_name(var_name)
        self._var_name_aerocom = None

        self.standard_name = None
        # Assume variables that have no unit specified in variables.ini are
        # unitless.
        self.units = '1'
        self.default_vert_code = None

        self.wavelength_nm = None
        self.dry_rh_max = 40
        self.dimensions = None
        self.minimum = self.VMIN_DEFAULT
        self.maximum = self.VMAX_DEFAULT

        self.description = None
        self.comments_and_purpose = None

        #wavelength tolerance in nm
        self.obs_wavelength_tol_nm = None

        self.scat_xlim = None
        self.scat_ylim = None
        self.scat_loglog = None
        self.scat_scale_factor = 1.0

        # settings for map plotting
        self.map_cmap = 'coolwarm'
        self.map_vmin = None
        self.map_vmax = None
        self.map_c_under = None
        self.map_c_over = 'r'
        self.map_cbar_levels = None
        self.map_cbar_ticks = None

        self._is_rate = False

        # imports default information and, on top, variable information (if
        # applicable)
        if init:
            self.parse_from_ini(self.var_name, cfg=cfg)

        self.update(**kwargs)
        if self.obs_wavelength_tol_nm is None:
            self.obs_wavelength_tol_nm = OBS_WAVELENGTH_TOL_NM

    @property
    def var_name_aerocom(self):
        """AeroCom variable name of the input variable"""
        vna = self._var_name_aerocom
        return self.var_name if vna is None else vna

    @property
    def var_name_input(self):
        """Input variable"""
        return self._var_name_input

    @property
    def is_3d(self):
        """True if str '3d' is contained in :attr:`var_name_input`"""
        return True if '3d' in self.var_name_input.lower() else False

    @property
    def is_wavelength_dependent(self):
        """Indicates whether this variable is wavelength dependent"""
        return True if self.wavelength_nm is not None else False

    @property
    def is_at_dry_conditions(self):
        """Indicate whether variable denotes dry conditions"""
        var_name = self.var_name_aerocom
        if var_name.startswith('dry'): # dry deposition
            return False
        return True if 'dry' in var_name else False

    @property
    def is_deposition(self):
        """
        Indicates whether input variables is a deposition rate

        Note
        ----
        This funtion only identifies wet and dry deposition based on the variable
        names, there might be other variables that are deposition variables but
        cannot be identified by this function.

        Parameters
        ----------
        var_name : str
            Name of variable to be checked

        Returns
        -------
        bool
            If True, then variable name denotes a deposition variables

        """
        var_name = self.var_name_aerocom
        if var_name.startswith(var_groups.drydep_startswith):
            return True
        elif var_name.startswith(var_groups.wetdep_startswith):
            return True
        elif var_name.startswith(var_groups.totdep_startswith):
            return True
        elif var_name in var_groups.dep_add_vars:
            return True
        return False

    @property
    def is_emission(self):
        """
        Indicates whether input variables is an emission rate

        Note
        ----
        This funtion only identifies wet and dry deposition based on the variable
        names, there might be other variables that are deposition variables but
        cannot be identified by this function.

        Parameters
        ----------
        var_name : str
            Name of variable to be checked

        Returns
        -------
        bool
            If True, then variable name denotes a deposition variables

        """
        var_name = self.var_name_aerocom
        if var_name.startswith(var_groups.emi_startswith):
            return True
        elif var_name in var_groups.emi_add_vars:
            return True
        return False

    @property
    def is_rate(self):
        """Indicates whether variable name is a rate

        Rates include e.g. deposition or emission rate variables but also
        precipitation

        Returns
        -------
        bool
            True if variable is rate, else False
        """
        if self.is_emission:
            return True
        elif self.is_deposition:
            return True
        elif self._is_rate:
            return True
        return False

    @property
    def is_alias(self):
        return True if self.var_name != self.var_name_aerocom else False

    @property
    def unit(self):
        """Unit of variable (old name, deprecated)"""
        from warnings import warn
        warn(DeprecationWarning('Attr. name unit in Variable '
                                'class is deprecated. Please '
                                'use units instead'))
        return self.units

    @property
    def plot_info(self):
        """Dictionary containing plot information"""
        d = {}
        for k in self.plot_info_keys:
            d[k] = self[k]
        return d

    def update(self, **kwargs):
        for key, val in kwargs.items():
            self[key] = val

    @property
    def has_unit(self):
        """Boolean specifying whether variable has unit"""
        return True if not self.units in (1, None) else False

    @property
    def lower_limit(self):
        """Old attribute name for :attr:`minimum` (following HTAP2 defs)"""
        logger.warning(DeprecationWarning('Old name for attribute minimum'))
        return self.minimum

    @property
    def upper_limit(self):
        """Old attribute name for :attr:`minimum` (following HTAP2 defs)"""
        logger.warning(DeprecationWarning('Old name for attribute minimum'))
        return self.maximum

    @property
    def unit_str(self):
        """string representation of unit"""
        if self.units is None:
            return ''
        else:
            return '[{}]'.format(self.units)

    @staticmethod
    def read_config():
        return parse_variables_ini()

    @property
    def var_name_info(self):

        return VarNameInfo(self.var_name)


    @property
    def aliases(self):
        """Alias variable names that are frequently found or used

        Returns
        -------
        list
            list containing valid aliases
        """
        return get_aliases(self.var_name)

    @property
    def long_name(self):
        """Wrapper for :attr:`description`"""
        return self.description

    def keys(self):
        return list(self.__dict__.keys())

    @staticmethod
    def _check_aliases(var_name):
        ap = parse_aliases_ini()
        aliases = _read_alias_ini(ap)
        if var_name in aliases:
            return aliases[var_name]
        return _check_alias_family(var_name, ap)

    def get_default_vert_code(self):
        """Get default vertical code for variable name"""
        if self.default_vert_code is not None:
            return self.default_vert_code
        try:
            return VarNameInfo(self.var_name_aerocom).get_default_vert_code()
        except ValueError:
            print_log.warning('default_vert_code not set for {} and '
                           'could also not be inferred'
                           .format(self.var_name_aerocom))
            return None

    def get_cmap(self):
        """
        Get cmap str for var

        Returns
        -------
        str

        """
        return self.map_cmap

    def _cmap_bins_from_vmin_vmax(self):
        """
        Calculate cmap discretisation bins from :attr:`vmin` and :attr:`vmax`

        Sets value of :attr:`map_cbar_levels`

        Raises
        ------
        AttributeError
             if :attr:`vmin` and :attr:`vmax` are not defined

        """
        if self.minimum == self.VMIN_DEFAULT or self.maximum == \
                self.VMAX_DEFAULT:
            raise AttributeError(f'need minimum and maximum to be specified '
                                 f'for variable {self.var_name} in '
                                 f'order to retrieve cmap_bins')
        self.map_cbar_levels = make_binlist(self.minimum, self.maximum)

    def get_cmap_bins(self, infer_if_missing=True):
        """
        Get cmap discretisation bins

        Parameters
        ----------
        infer_if_missing : bool
            if True and :attr:`map_cbar_levels` is not defined, try to infer
            using :func:`_cmap_bins_from_vmin_vmax`.

        Raises
        ------
        AttributeError
             if unavailable

        Returns
        -------
        list
            levels

        """
        if self.map_cbar_levels is None:
            if infer_if_missing:
                self._cmap_bins_from_vmin_vmax()
            else:
                raise AttributeError(f'map_cbar_levels is not defined for '
                                     f'variable {self.var_name}')
        return self.map_cbar_levels

    def parse_from_ini(self, var_name=None, cfg=None):
        """Import information about default region

        Parameters
        ----------
        var_name : str
            variable name
        var_name_alt : str
            alternative variable name that is used if variable name is not
            available
        cfg : ConfigParser
            open config parser object

        Returns
        -------
        bool
            True, if default could be loaded, False if not
        """
        if cfg is None:
            cfg = self.read_config()
        elif not isinstance(cfg, ConfigParser):
            raise ValueError(f'invalid input for cfg, need config parser got '
                             f'{type(cfg)}')
        if not var_name in cfg:
            try:
                var_name = self._check_aliases(var_name)
            except VariableDefinitionError:
                logger.info('Unknown input variable {}'.format(var_name))
                return
            self._var_name_aerocom = var_name

        var_info = cfg[var_name]
        # this variable should import settings from another variable
        if 'use' in var_info:
            use = var_info['use']
            if not use in cfg:
                raise VariableDefinitionError('Input variable {} depends on {} '
                                              'which is not available in '
                                              'variables.ini.'
                                              .format(var_name, use))
            self.parse_from_ini(use, cfg)

        for key, val in var_info.items():
            if key in self.ALT_NAMES:
                key = self.ALT_NAMES[key]
            self._add(key, val)

    def _add(self, key, val):
        if key in self._TYPE_CONV:
            try:
                val = self._TYPE_CONV[key](val)
            except:
                pass
        elif key == 'units' and val == 'None':
            val = '1'
        if val == 'None':
            val = None
        self[key] = val

    def __setitem__(self, key, val):
        self.__dict__[key] = val

    def __getitem__(self, key):
        return self.__dict__[key]

    def __repr__(self):
       return ("{}\nstandard_name: {}; Unit: {}"
               .format(self.var_name, self.standard_name,
                       self.units))

    def __eq__(self, other):
        if isinstance(other, str):
            other = Variable(other)
        elif not isinstance(other, Variable):
            raise TypeError('Can only compare with str or other Variable instance')
        return True if other.var_name_aerocom == self.var_name_aerocom else False

    def __str__(self):
        head = "Pyaerocom {}".format(type(self).__name__)
        s = "\n{}\n{}".format(head, len(head)*"-")

        plot_s = '\nPlotting settings\n......................'

        for k, v in self.__dict__.items():
            if k in self.plot_info_keys:
                if v is None:
                    continue
                if isinstance(v, dict):
                    plot_s += "\n{} (dict)".format(k)
                    plot_s += dict_to_str(v, indent=3,
                                         ignore_null=True)
                elif isinstance(v, list):
                    plot_s += "\n{} (list, {} items)".format(k, len(v))
                    plot_s += list_to_shortstr(v)
                else:
                    plot_s += "\n%s: %s" %(k,v)
            else:
                if isinstance(v, dict):
                    s += "\n{} (dict)".format(k)
                    s += dict_to_str(v, indent=3, ignore_null=True)
                elif isinstance(v, list):
                    s += "\n{} (list, {} items)".format(k, len(v))
                    s += list_to_shortstr(v)
                else:
                    s += "\n%s: %s" %(k,v)

        s += plot_s
        return s